# -*- coding: utf-8 -*-
"""
Created on Fri Nov  5 01:34:00 2021

@author: yrc2
"""
import biosteam as bst
import biorefineries.oilcane as oc
from biosteam.utils import CABBI_colors, colors
from thermosteam.utils import set_figure_size, set_font
from thermosteam.units_of_measure import format_units
from colorpalette import Palette
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from warnings import warn
import numpy as np
import pandas as pd
from matplotlib.lines import Line2D
from matplotlib.gridspec import GridSpec
from . import _variable_mockups as variables
from ._variable_mockups import (
    tea_monte_carlo_metric_mockups,
    tea_monte_carlo_derivative_metric_mockups,
    lca_monte_carlo_metric_mockups, 
    lca_monte_carlo_derivative_metric_mockups,
    MFPP, TCI, electricity_production, natural_gas_consumption,
    ethanol_production, biodiesel_production,
    GWP_ethanol, GWP_biodiesel, GWP_electricity,
    GWP_ethanol_allocation, GWP_biodiesel_allocation,
    GWP_economic, MFPP_derivative, 
    TCI_derivative, 
    ethanol_production_derivative,
    biodiesel_production_derivative,
    electricity_production_derivative,
    natural_gas_consumption_derivative,
    GWP_ethanol_derivative,
)
from ._load_data import (
    images_folder,
    get_monte_carlo,
    spearman_file,
)
import os
from._parse_configuration import format_name
from math import log10, floor
from collections import Iterable

__all__ = (
    'plot_all',
    'plot_montecarlo_main_manuscript',
    'plot_breakdowns',
    'plot_montecarlo_feedstock_comparison',
    'plot_montecarlo_configuration_comparison',
    'plot_montecarlo_agile_comparison',
    'plot_montecarlo_derivative',
    'plot_montecarlo_absolute',
    'plot_spearman_tea',
    'plot_spearman_lca',
    'plot_monte_carlo_across_coordinate',
    'monte_carlo_box_plot',
    'plot_monte_carlo',
    'plot_spearman',
    'plot_configuration_breakdown',
    'plot_TCI_areas_across_oil_content',
    'plot_heatmap_comparison',
    'plot_feedstock_conventional_comparison_kde',
    'plot_feedstock_cellulosic_comparison_kde',
    'plot_configuration_comparison_kde',
    'plot_open_comparison_kde',
    'monte_carlo_results',
    'montecarlo_results_feedstock_comparison',
    'montecarlo_results_configuration_comparison',
    'montecarlo_results_agile_comparison',
    'plot_feedstock_comparison_kde',
    'plot_crude_configuration_comparison_kde',
    'plot_agile_comparison_kde',
)

area_colors = {
    'Feedstock handling': CABBI_colors.teal, 
    'Juicing': CABBI_colors.green_dirty,
    'EtOH prod.': CABBI_colors.blue,
    'Ethanol production': CABBI_colors.blue,
    'Oil ext.': CABBI_colors.brown,
    'Oil extraction': CABBI_colors.brown,
    'Biod. prod.': CABBI_colors.orange,
    'Biodiesel production': CABBI_colors.orange,
    'Pretreatment': CABBI_colors.green,
    'Wastewater treatment': colors.purple,
    'CH&P': CABBI_colors.yellow,
    'Co-Heat and Power': CABBI_colors.yellow,
    'Utilities': colors.red,
    'Storage': CABBI_colors.grey,
    'HXN': colors.orange,
    'Heat exchanger network': colors.orange,
}

area_hatches = {
    'Feedstock handling': 'x', 
    'Juicing': '-',
    'EtOH prod.': '/',
    'Ethanol production': '/',
    'Oil ext.': '\\',
    'Oil extraction': '\\',
    'Biod. prod.': '/|',
    'Biodiesel production': '/|',
    'Pretreatment': '//',
    'Wastewater treatment': r'\\',
    'CH&P': '',
    'Co-Heat and Power': '',
    'Utilities': '\\|',
    'Storage': '',
    'HXN': '+',
    'Heat exchanger network': '+',
}

for i in area_colors: area_colors[i] = area_colors[i].tint(20)
palette = Palette(**area_colors)
letter_color = colors.neutral.shade(25).RGBn
GWP_units_gal = '$\\mathrm{kg} \\cdot \\mathrm{CO}_{2}\\mathrm{eq} \\cdot \\mathrm{gal}^{-1}$'
CABBI_colors.orange_hatch = CABBI_colors.orange.copy(hatch='////')

def roundsigfigs(x, nsigfigs=2):
    if isinstance(x, Iterable):
        n = nsigfigs - int(floor(log10(abs(x[1])))) - 1 if abs(x[1]) > 1e-12 else 0.
        try:
            value = np.round(x, n)
        except:
            return np.array(x, dtype=int)
        if (np.array(value, int) == value).all():
            return np.array(value, int)
        else:
            return value
    else:
        n = nsigfigs - int(floor(log10(abs(x)))) - 1 if abs(x) > 1e-12 else 0.
        try:
            value = round(x, n)
        except:
            return int(x)
        if int(value) == value:
            return int(value)
        else:
            return value
    
ethanol_over_biodiesel = bst.MockVariable('Ethanol over biodiesel', 'gal/ton', 'Biorefinery')
GWP_ethanol_displacement = variables.GWP_ethanol_displacement
production = (ethanol_production, biodiesel_production)

mc_metric_settings = {
    'MFPP': (MFPP, f"MFPP\n[{format_units('USD/ton')}]", None),
    'TCI': (TCI, f"TCI\n[{format_units('10^6*USD')}]", None),
    'production': (production, f"Production\n[{format_units('gal/ton')}]", None),
    'electricity_production': (electricity_production, f"Elec. prod.\n[{format_units('kWhr/ton')}]", None),
    'natural_gas_consumption': (natural_gas_consumption, f"NG cons.\n[{format_units('cf/ton')}]", None),
    'GWP_ethanol_displacement': (GWP_ethanol_displacement, "GWP$_{\\mathrm{displacement}}$" f"\n[{GWP_units_gal}]", None),
    'GWP_economic': ((GWP_ethanol, GWP_biodiesel), "GWP$_{\\mathrm{economic}}$" f"\n[{GWP_units_gal}]", None),
    'GWP_energy': ((GWP_ethanol_allocation, GWP_biodiesel_allocation), "GWP$_{\\mathrm{energy}}$" f"\n[{GWP_units_gal}]", None),
}

mc_comparison_settings = {
    'MFPP': (MFPP, r"$\Delta$" + f"MFPP\n[{format_units('USD/ton')}]", None),
    'TCI': (TCI, r"$\Delta$" + f"TCI\n[{format_units('10^6*USD')}]", None),
    'production': (production, r"$\Delta$" + f"Production\n[{format_units('gal/ton')}]", None),
    'electricity_production': (electricity_production, r"$\Delta$" + f"Elec. prod.\n[{format_units('kWhr/ton')}]", None),
    'natural_gas_consumption': (natural_gas_consumption, r"$\Delta$" + f"NG cons.\n[{format_units('cf/ton')}]", None),
    'GWP_ethanol_displacement': (GWP_ethanol_displacement, r"$\Delta$" + "GWP$_{\\mathrm{displacement}}$" f"\n[{GWP_units_gal}]", None),
    'GWP_economic': (GWP_ethanol, r"$\Delta$" + "GWP$_{\\mathrm{economic}}$" f"\n[{GWP_units_gal}]", None),
    'GWP_energy': (GWP_ethanol_allocation, r"$\Delta$" + "GWP$_{\\mathrm{energy}}$" f"\n[{GWP_units_gal}]", None),
    'GWP_property_allocation': ((GWP_ethanol, GWP_ethanol_allocation), r"$\Delta$" + f"GWP\n[{GWP_units_gal}]", None),
}

mc_derivative_metric_settings = {
    'MFPP': (MFPP_derivative, r"$\Delta$" + format_units(r"MFPP/OC").replace('cdot', r'cdot \Delta') + f"\n[{format_units('USD/ton')}]", None),
    'TCI': (TCI_derivative,  r"$\Delta$" + format_units(r"TCI/OC").replace('cdot', r'cdot \Delta') + f"\n[{format_units('10^6*USD')}]", None),
    'production': ((ethanol_production_derivative, biodiesel_production_derivative), r"$\Delta$" + format_units(r"Prod./OC").replace('cdot', r'cdot \Delta') + f"\n[{format_units('gal/ton')}]", None),
    'electricity_production': (electricity_production_derivative, r"$\Delta$" + format_units(r"EP/OC").replace('cdot', r'cdot \Delta') + f"\n[{format_units('kWhr/ton')}]", None),
    'natural_gas_consumption': (natural_gas_consumption_derivative, r"$\Delta$" + format_units(r"NGC/OC").replace('cdot', r'cdot \Delta') + f"\n[{format_units('cf/ton')}]", None),
    'GWP_economic': (GWP_ethanol_derivative, r"$\Delta$" + r"GWP $\cdot \Delta \mathrm{OC}^{-1}$" f"\n[{GWP_units_gal.replace('kg','g')}]", 1000),
}

# for dct in (mc_metric_settings, mc_comparison_settings, mc_derivative_metric_settings):
#     for i, j in dct.copy().items(): dct[j[0]] = j
    
# %% Plots for publication

def plot_all():
    # plot_montecarlo_main_manuscript()
    plot_montecarlo_absolute()
    plot_spearman_tea()
    plot_spearman_lca()
    plot_breakdowns()

def plot_montecarlo_main_manuscript():
    set_font(size=8)
    set_figure_size(aspect_ratio=0.85)
    fig = plt.figure()
    everything = GridSpec(4, 3, fig, hspace=1.5, wspace=0.7,
                          top=0.90, bottom=0.05,
                          left=0.11, right=0.97)
    
    def spec2axes(spec, x, y, hspace=0, wspace=0.7, **kwargs):
        subspec = spec.subgridspec(x, y, hspace=hspace, wspace=wspace, **kwargs)
        return np.array([[fig.add_subplot(subspec[i, j]) for j in range(y)] for i in range(x)], object)
    
    gs_feedstock_comparison = everything[:2, :]
    gs_configuration_comparison = everything[2:, :2]
    gs_agile_comparison = everything[2:, 2]
    axes_feedstock_comparison = spec2axes(gs_feedstock_comparison, 2, 3)
    axes_configuration_comparison = spec2axes(gs_configuration_comparison, 2, 2)
    axes_agile_comparison = spec2axes(gs_agile_comparison, 2, 1)
    plot_montecarlo_feedstock_comparison(axes_feedstock_comparison, letters='ABCDEFG')
    plot_montecarlo_configuration_comparison(axes_configuration_comparison, letters='ABCDEFG')
    plot_montecarlo_agile_comparison(axes_agile_comparison, letters='ABCDEFG')
    
    def add_title(gs, title):
        ax =  fig.add_subplot(gs)
        ax._frameon = False
        ax.xaxis.set_visible(False)
        ax.yaxis.set_visible(False)
        ax.set_title(
            title, color=letter_color,
            horizontalalignment='center',verticalalignment='center',
            fontsize=12, fontweight='bold', y=1.1
        )
    add_title(gs_feedstock_comparison, '(I) Impact of opting to process oilcane over sugarcane')
    add_title(gs_configuration_comparison, '(II) Impact of cellulosic ethanol integration')
    add_title(gs_agile_comparison, '(III) Impact of\noilsorghum\nintegration')
    plt.show()
    for i in ('svg', 'png'):
        file = os.path.join(images_folder, f'montecarlo_main_manuscript.{i}')
        plt.savefig(file, transparent=True)

def plot_montecarlo_feedstock_comparison(axes_box=None, letters=None, 
                                         single_column=True):
    if single_column:
        width = 'half'
        aspect_ratio = 2.25
        ncols = 1
        left = 0.255
        bottom = 0.05
    else:
        width = None
        aspect_ratio = 0.75
        left = 0.105
        bottom = 0.12
        ncols = 3
    if axes_box is None:
        set_font(size=8)
        set_figure_size(width=width, aspect_ratio=aspect_ratio)       
    fig, axes = plot_monte_carlo(
        derivative=False, absolute=False, comparison=True,
        tickmarks=None, agile=False, ncols=ncols, axes_box=axes_box,
        labels=[
            'Conventional',
            'Cellulosic',
            # 'Conventional',
            # 'Cellulosic',
        ],
        comparison_names=['O1 - S1', 'O2 - S2'],
        metrics = ['MFPP', 'TCI', 'production', 'GWP_property_allocation', 
                   'natural_gas_consumption', 'electricity_production'],
        color_wheel = CABBI_colors.wheel([
            'blue_light', 'green_dirty', 'orange', 'green', 
            'orange', 'orange_hatch', 'grey', 'brown',
        ])
    )
    for ax, letter in zip(axes, 'ABCDEFGH' if letters is None else letters):
        plt.sca(ax)
        ylb, yub = plt.ylim()
        plt.text(1.65, ylb + (yub - ylb) * 0.90, letter, color=letter_color,
                 horizontalalignment='center',verticalalignment='center',
                 fontsize=12, fontweight='bold')
        # if axes_box is None and letter in 'DH':
        #     x = 0.5
        #     plt.text(x, ylb - (yub - ylb) * 0.3, 
        #              'Impact of processing\noilcane over sugarcane', 
        #              horizontalalignment='center',verticalalignment='center',
        #              fontsize=8)
    if axes_box is None:
        plt.subplots_adjust(right=0.96, left=left, wspace=0.38, top=0.98, bottom=bottom)
        for i in ('svg', 'png'):
            file = os.path.join(images_folder, f'montecarlo_feedstock_comparison.{i}')
            plt.savefig(file, transparent=True)

    
def plot_montecarlo_configuration_comparison(axes_box=None, letters=None,
                                             single_column=True):
    if single_column:
        width = 'half'
        aspect_ratio = 2.25
        ncols = 1
        left = 0.255
        bottom = 0.05
        x = 1.65
        metrics= ['MFPP', 'TCI', 'production', 'GWP_property_allocation',
                  'natural_gas_consumption', 'electricity_production']
    else:
        width = None
        aspect_ratio = 0.75
        left = 0.105
        bottom = 0.12
        ncols = 2
        x = 0.58
        metrics= ['MFPP', 'TCI', 'production', 'GWP_property_allocation']
    if axes_box is None:
        set_font(size=8)
        set_figure_size(width=width, aspect_ratio=aspect_ratio)
    fig, axes = plot_monte_carlo(
        derivative=False, absolute=False, comparison=True,
        tickmarks=None, agile=False, ncols=ncols, axes_box=axes_box,
        labels=[
            'Oilcane',
            # 'Sugarcane',
        ],
        comparison_names=[
            'O2 - O1', 
            # 'S2 - S1'
        ],
        metrics=metrics,
        color_wheel = CABBI_colors.wheel([
            'blue_light', 'green_dirty', 'orange', 'green', 
            'orange', 'orange_hatch', 
        ])
    )
    for ax, letter in zip(axes, 'ABCDEF' if letters is None else letters):
        plt.sca(ax)
        ylb, yub = plt.ylim()
        plt.text(x, ylb + (yub - ylb) * 0.90, letter, color=letter_color,
                 horizontalalignment='center',verticalalignment='center',
                 fontsize=12, fontweight='bold')
    if axes_box is None:
        plt.subplots_adjust(right=0.96, left=left, wspace=0.38, top=0.98, bottom=bottom)
        for i in ('svg', 'png'):
            file = os.path.join(images_folder, f'montecarlo_configuration_comparison.{i}')
            plt.savefig(file, transparent=True)
    
def plot_montecarlo_agile_comparison(axes_box=None, letters=None):
    if axes_box is None:
        set_font(size=8)
        set_figure_size(width=3.3071, aspect_ratio=1.0)
    fig, axes = plot_monte_carlo(
        derivative=False, absolute=False, comparison=True,
        tickmarks=None, agile_only=True, ncols=1,
        labels=[
            'Conventional',
            'Cellulosic'
        ],
        metrics=['MFPP', 'TCI'],
        axes_box=axes_box,
    )
    for ax, letter in zip(axes, 'AB'  if letters is None else letters):
        plt.sca(ax)
        ylb, yub = plt.ylim()
        plt.text(1.65, ylb + (yub - ylb) * 0.90, letter, color=letter_color,
                 horizontalalignment='center',verticalalignment='center',
                 fontsize=12, fontweight='bold')
        if axes_box is None and letter == 'B':
            plt.text(0.5, ylb - (yub - ylb) * 0.25, 
                      'Impact of integrating oilsorghum\nat an agile oilcane biorefinery', 
                      horizontalalignment='center',verticalalignment='center',
                      fontsize=8)
    if axes_box is None:
        plt.subplots_adjust(right=0.9, left=0.2, wspace=0.5, top=0.98, bottom=0.15)
        for i in ('svg', 'png'):
            file = os.path.join(images_folder, f'montecarlo_agile_comparison.{i}')
            plt.savefig(file, transparent=True)
    
def plot_montecarlo_derivative():
    set_font(size=8)
    set_figure_size(
        aspect_ratio=0.5,
        # width=3.3071, aspect_ratio=1.85
    )
    fig, axes = plot_monte_carlo(
        derivative=True, absolute=True, 
        comparison=False, agile=False,
        ncols=3,
        # tickmarks=np.array([
        #     [-3, -2, -1, 0, 1, 2, 3, 4, 5],
        #     [-9, -6, -3,  0, 3, 6, 9, 12, 15],
        #     [-2.0, -1.5, -1.0, -0.5, 0, 0.5, 1.0, 1.5, 2],
        #     [-16, -8, 0, 8, 16, 24, 32, 40, 48],
        #     [-400, -300, -200, -100, 0, 100, 200, 300, 400],
        #     [-300, -225, -150, -75, 0, 75, 150, 225, 300]
        # ], dtype=object),
        labels=['Conventional', 'Cellulosic'],
        color_wheel = CABBI_colors.wheel([
            'blue_light', 'green_dirty', 'orange', 'green', 'grey', 'brown',
            'orange', 
        ])
    )
    for ax, letter in zip(axes, 'ABCDEFGH'):
        plt.sca(ax)
        ylb, yub = plt.ylim()
        plt.text(1.65, ylb + (yub - ylb) * 0.90, letter, color=letter_color,
                 horizontalalignment='center',verticalalignment='center',
                 fontsize=12, fontweight='bold')
    plt.subplots_adjust(
        hspace=0, wspace=0.7,
        top=0.95, bottom=0.1,
        left=0.12, right=0.96
    )
    for i in ('svg', 'png'):
        file = os.path.join(images_folder, f'montecarlo_derivative.{i}')
        plt.savefig(file, transparent=True)

def plot_montecarlo_absolute():
    set_font(size=8)
    set_figure_size(aspect_ratio=1.05)
    fig, axes = plot_monte_carlo(
        absolute=True, comparison=False, ncols=2,
        expand=0.1, 
        labels=['Sugarcane\nConventional', 'Oilcane\nConventional',
                'Sugarcane\nCellulosic', 'Oilcane\nCellulosic',
                'Sugarcane/Sorghum\nConventional', 'Oilcane/Oilsorghum\nConventional',
                'Sugarcane\Sorghum\nCellulosic', 'Oilcane/Oilsorghum\nCellulosic'],
        xrot=90,
        color_wheel = CABBI_colors.wheel([
            'blue_light', 'green_dirty', 'orange', 'green', 'grey', 'brown',
            'orange', 'orange', 'green', 'orange', 'green',
        ])
    )
    for ax, letter in zip(axes, 'ABCDEFGHIJ'):
        plt.sca(ax)
        ylb, yub = plt.ylim()
        plt.text(7.8, ylb + (yub - ylb) * 0.92, letter, color=letter_color,
                 horizontalalignment='center',verticalalignment='center',
                 fontsize=12, fontweight='bold')
    plt.subplots_adjust(left=0.12, right=0.95, wspace=0.40, top=0.98, bottom=0.2)
    for i in ('svg', 'png'):
        file = os.path.join(images_folder, f'montecarlo_absolute.{i}')
        plt.savefig(file, transparent=True)
    
def plot_spearman_tea():
    set_font(size=8)
    set_figure_size(aspect_ratio=0.80)
    plot_spearman(
        configurations=[
            'O1', 'O1*',
            'O2', 'O2*',
        ],
        labels=[
            'Conventional', 'Agile-Conventional',
            'Cellulosic', 'Agile-Cellulosic',
        ],
        cutoff=0.025,
        kind='TEA',
    )
    plt.subplots_adjust(left=0.45, right=0.975, top=0.98, bottom=0.08)
    for i in ('svg', 'png'):
        file = os.path.join(images_folder, f'spearman_tea.{i}')
        plt.savefig(file, transparent=True)

def plot_spearman_lca():
    set_font(size=8)
    set_figure_size(aspect_ratio=0.65)
    plot_spearman(
        configurations=[
            'O1', 'O1*',
            'O2', 'O2*',
        ],
        labels=[
            'Conventional', 'Agile-Conventional',
            'Cellulosic', 'Agile-Cellulosic',
        ],
        cutoff=0.03,
        kind='LCA',
    )
    plt.subplots_adjust(left=0.45, right=0.975, top=0.98, bottom=0.10)
    for i in ('svg', 'png'):
        file = os.path.join(images_folder, f'spearman_lca.{i}')
        plt.savefig(file, transparent=True)

def plot_breakdowns():
    set_font(size=8)
    set_figure_size(aspect_ratio=0.68)
    fig, axes = plt.subplots(nrows=1, ncols=2)
    plt.sca(axes[0])
    plot_configuration_breakdown('O1', ax=axes[0], legend=False)
    plt.sca(axes[1])
    plot_configuration_breakdown('O2', ax=axes[1], legend=True)
    yticks = axes[1].get_yticks()
    plt.yticks(yticks, ['']*len(yticks))
    plt.ylabel('')
    plt.subplots_adjust(left=0.09, right=0.96, wspace=0., top=0.84, bottom=0.31)
    for ax, letter in zip(axes, ['(A) Conventional', '(B) Cellulosic']):
        plt.sca(ax)
        ylb, yub = plt.ylim()
        xlb, xub = plt.xlim()
        plt.text((xlb + xub) * 0.5, ylb + (yub - ylb) * 1.2, letter, color=letter_color,
                  horizontalalignment='center',verticalalignment='center',
                  fontsize=12, fontweight='bold')
    for i in ('svg', 'png'):
        file = os.path.join(images_folder, f'breakdowns.{i}')
        plt.savefig(file, transparent=True)

# %% Monte carlo values for manuscript

def montecarlo_results_feedstock_comparison():
    return montecarlo_results_short(
        metrics = [
            MFPP, TCI, ethanol_production, biodiesel_production, 
            electricity_production, natural_gas_consumption, GWP_ethanol_displacement, 
            GWP_ethanol, GWP_ethanol_allocation, GWP_ethanol, GWP_ethanol_allocation
        ],
        names = [
            'O1 - S1',
            'O2 - S2',
        ],
    )
    
def montecarlo_results_configuration_comparison():
    return montecarlo_results_short(
        metrics = [
            MFPP, TCI, ethanol_production, biodiesel_production, 
            electricity_production, natural_gas_consumption, GWP_ethanol_displacement, 
            GWP_ethanol, GWP_ethanol_allocation, GWP_ethanol, GWP_ethanol_allocation
        ],
        names = [
            'O2 - O1',
        ],
    )['O2 - O1']

def montecarlo_results_agile_comparison():
    return montecarlo_results_short(
        metrics = [
            MFPP, TCI, ethanol_production, biodiesel_production, 
            electricity_production, natural_gas_consumption, GWP_ethanol_displacement, 
            GWP_ethanol, GWP_ethanol_allocation, GWP_ethanol, GWP_ethanol_allocation
        ],
        names = [
            'O1* - O1',
            'O2* - O2',
        ],
    )

def montecarlo_results_short(names, metrics):
    results = {}
    for name in names:
        df = get_monte_carlo(name)
        results[name] = dct = {}
        for metric in metrics:
            index = metric.index
            data = df[index].values
            q05, q50, q95 = roundsigfigs(np.percentile(data, [5, 50, 95], axis=0), 3)
            key = get_monte_carlo_key(index, dct, False)
            dct[key] = f"{q50} [{q05}, {q95}]"
    return results

# %% Heatmap

def get_fraction_in_same_direction(data, direction):
    return (direction * data >= 0.).sum(axis=0) / data.size

def get_median(data):
    return roundsigfigs(np.percentile(data, 50, axis=0))

def plot_heatmap_comparison(comparison_names=None, xlabels=None):
    if comparison_names is None: comparison_names = oc.comparison_names
    columns = comparison_names
    if xlabels is None: xlabels = [format_name(i).replace(' ', '') for i in comparison_names]
    def get_data(metric, name):
        df = get_monte_carlo(name, metric)
        values = df.values
        return values
    
    GWP_economic, GWP_ethanol, GWP_biodiesel, GWP_electricity, GWP_crude_glycerol, = lca_monte_carlo_metric_mockups
    MFPP, TCI, ethanol_production, biodiesel_production, electricity_production, natural_gas_consumption = tea_monte_carlo_metric_mockups
    GWP_ethanol_displacement = variables.GWP_ethanol_displacement
    GWP_ethanol_allocation = variables.GWP_ethanol_allocation
    rows = [
        MFPP, 
        TCI, 
        ethanol_production, 
        biodiesel_production,
        electricity_production,
        natural_gas_consumption,
        GWP_ethanol_displacement,
        GWP_ethanol_allocation,
        GWP_ethanol, # economic
    ]
    ylabels = [
        f"MFPP\n[{format_units('USD/ton')}]",
        f"TCI\n[{format_units('10^6*USD')}]",
        f"Ethanol production\n[{format_units('gal/ton')}]",
        f"Biodiesel production\n[{format_units('gal/ton')}]",
        f"Elec. prod.\n[{format_units('kWhr/ton')}]",
        f"NG cons.\n[{format_units('cf/ton')}]",
        "GWP$_{\\mathrm{displacement}}$" f"\n[{GWP_units_gal}]",
        "GWP$_{\\mathrm{energy}}$" f"\n[{GWP_units_gal}]",
        "GWP$_{\\mathrm{economic}}$" f"\n[{GWP_units_gal}]",
    ]
    N_rows = len(rows)
    N_cols = len(comparison_names)
    data = np.zeros([N_rows, N_cols], dtype=object)
    data[:] = [[get_data(i, j) for j in columns] for i in rows]
    medians = np.zeros_like(data, dtype=float)
    fractions = medians.copy()
    for i in range(N_rows):
        for j in range(N_cols):
            medians[i, j] = x = get_median(data[i, j])
            fractions[i, j] = get_fraction_in_same_direction(data[i, j], 1 if x > 0 else -1)
            
    fig, ax = plt.subplots()
    mbar = bst.plots.MetricBar(
        'Fraction in the same direction [%]', ticks=[-100, -75, -50, -25, 0, 25, 50, 75, 100],
        cmap=plt.cm.get_cmap('RdYlGn')
    )
    im, cbar = bst.plots.plot_heatmap(
        100 * fractions, vmin=0, vmax=100, ax=ax, cell_labels=medians,
        metric_bar=mbar, xlabels=xlabels, ylabels=ylabels,
    )
    cbar.ax.set_ylabel(mbar.title, rotation=-90, va="bottom")
    plt.sca(ax)
    ax.spines[:].set_visible(False)
    plt.grid(True, 'major', 'both', lw=1, color='w', ls='-')

# %% KDE


def plot_kde(name, metrics=(GWP_ethanol, MFPP), xticks=None, yticks=None,
             xbox_kwargs=None, ybox_kwargs=None, top_left='',
             top_right='Tradeoff', bottom_left='Tradeoff',
             bottom_right=''):
    set_font(size=8)
    set_figure_size(width='half', aspect_ratio=1.20)
    Xi, Yi = [i.index for i in metrics]
    df = oc.get_monte_carlo(name, metrics)
    y = df[Yi].values
    x = df[Xi].values
    # print(df[Xi].min(), df[Xi].max())
    # print(df[Yi].min(), df[Yi].max())
    ax = bst.plots.plot_kde(
        y=y, x=x, xticks=xticks, yticks=yticks,
        xticklabels=True, yticklabels=True,
        xbox_kwargs=xbox_kwargs or dict(light=CABBI_colors.orange.RGBn, dark=CABBI_colors.orange.shade(60).RGBn),
        ybox_kwargs=ybox_kwargs or dict(light=CABBI_colors.blue.RGBn, dark=CABBI_colors.blue.shade(60).RGBn),
    )
    plt.sca(ax)
    sX, sY = [mc_comparison_settings[i] for i in metrics]
    _, xlabel, _ = sX
    _, ylabel, _ = sY
    plt.xlabel(xlabel.replace('\n', ' '))
    plt.ylabel(ylabel.replace('\n', ' '))
    bst.plots.plot_quadrants()
    xlb, xub = plt.xlim()
    ylb, yub = plt.ylim()
    xpos = lambda x: xlb + (xub - xlb) * x
    # xlpos = lambda x: xlb * (1 - x)
    ypos = lambda y: ylb + (yub - ylb) * y
    y_mt_0 = y > 0
    y_lt_0 = y < 0
    x_mt_0 = x > 0
    x_lt_0 = x < 0
    xleft = 0.02
    xright = 0.98
    ytop = 0.94
    ybottom = 0.02
    if yub > 0. and xlb < 0.:
        if top_left.endswith('()'):
            p = (y_mt_0 & x_lt_0).sum() / y.size
            top_left = f"{p:.0%} {top_left.strip('()')}"
        plt.text(xpos(xleft), ypos(ytop), top_left, color=CABBI_colors.teal.shade(50).RGBn,
                 horizontalalignment='left', verticalalignment='top',
                 fontsize=10, fontweight='bold', zorder=10)
    if ylb < 0. and xlb < 0.:
        if bottom_left.endswith('()'):
            p = (y_lt_0 & x_lt_0).sum() / y.size
            bottom_left = f"{p:.0%} {bottom_left.strip('()')}"
        plt.text(xpos(xleft), ypos(ybottom), bottom_left, color=CABBI_colors.grey.shade(75).RGBn,
                 horizontalalignment='left', verticalalignment='bottom',
                 fontsize=10, fontweight='bold', zorder=10)
    if yub > 0. and xub > 0.:
        if top_right.endswith('()'):
            p = (y_mt_0 & x_mt_0).sum() / y.size
            top_right = f"{p:.0%} {top_right.strip('()')}"
        plt.text(xpos(xright), ypos(ytop), top_right, color=CABBI_colors.grey.shade(75).RGBn,
                 horizontalalignment='right', verticalalignment='top',
                 fontsize=10, fontweight='bold', zorder=10)
    if ylb < 0. and xub > 0.:
        if bottom_right.endswith('()'):
            p = (y_lt_0 & x_mt_0).sum() / y.size
            bottom_right = f"{p:.0%} {bottom_right.strip('()')}"
        plt.text(xpos(xright), ypos(ybottom), bottom_right, color=colors.red.shade(50).RGBn,
                 horizontalalignment='right', verticalalignment='bottom',
                 fontsize=10, fontweight='bold', zorder=10)
    plt.subplots_adjust(
        hspace=0.05, wspace=0.05,
        top=0.98, bottom=0.15,
        left=0.15, right=0.98,
    )

def plot_kde_2d(name, metrics=(GWP_ethanol, MFPP), xticks=None, yticks=None,
                top_left='', top_right='Tradeoff', bottom_left='Tradeoff',
                bottom_right='', ybox_kwargs=None, titles=None):
    set_font(size=8)
    set_figure_size(aspect_ratio=0.65)
    if isinstance(name, str): name = (name,)
    Xi, Yi = [i.index for i in metrics]
    dfs = [oc.get_monte_carlo(i, metrics) for i in name]
    axes = bst.plots.plot_kde_2d(
        ys=np.array([[df[Yi] for df in dfs]]),
        xs=np.array([[df[Xi] for df in dfs]]), 
        xticks=xticks, yticks=yticks,
        xticklabels=[True, True], yticklabels=[True, True],
        xbox_kwargs=2*[dict(light=CABBI_colors.orange.RGBn, dark=CABBI_colors.orange.shade(60).RGBn)],
        ybox_kwargs=[ybox_kwargs or dict(light=CABBI_colors.blue.RGBn, dark=CABBI_colors.blue.shade(60).RGBn)],
    )
    M, N = axes.shape
    xleft = 0.02
    xright = 0.98
    ytop = 0.94
    ybottom = 0.02
    for i in range(M):
        for j in range(N):
            ax = axes[i, j]
            plt.sca(ax)
            sX, sY = [mc_comparison_settings[i] for i in metrics]
            _, xlabel, _ = sX
            _, ylabel, _ = sY
            if i == M - 1: plt.xlabel(xlabel.replace('\n', ' '))
            if j == 0: plt.ylabel(ylabel.replace('\n', ' '))
            bst.plots.plot_quadrants()
            xlb, xub = plt.xlim()
            ylb, yub = plt.ylim()
            xpos = lambda x: xlb + (xub - xlb) * x
            # xlpos = lambda x: xlb * (1 - x)
            ypos = lambda y: ylb + (yub - ylb) * y
            df = dfs[j]
            x = df[Xi]
            y = df[Yi]
            y_mt_0 = y > 0
            y_lt_0 = y < 0
            x_mt_0 = x > 0
            x_lt_0 = x < 0
            if yub > 0. and xlb < 0. and top_left:
                if top_left.endswith('()'):
                    p = (y_mt_0 & x_lt_0).sum() / y.size
                    top_left = f"{p:.0%} {top_left.strip('()')}"
                    replacement = '()'
                else:
                    replacement = None
                plt.text(xpos(xleft), ypos(ytop), top_left, color=CABBI_colors.teal.shade(50).RGBn,
                         horizontalalignment='left', verticalalignment='top',
                         fontsize=10, fontweight='bold', zorder=10)
                top_left = replacement
            if ylb < 0. and xlb < 0. and bottom_left:
                if bottom_left.endswith('()'):
                    p = (y_lt_0 & x_lt_0).sum() / y.size
                    bottom_left = f"{p:.0%} {bottom_left.strip('()')}"
                    replacement = '()'
                else:
                    replacement = None
                plt.text(xpos(xleft), ypos(ybottom), bottom_left, color=CABBI_colors.grey.shade(75).RGBn,
                         horizontalalignment='left', verticalalignment='bottom',
                         fontsize=10, fontweight='bold', zorder=10)
                bottom_left = replacement
            if yub > 0. and xub > 0. and top_right:
                if top_right.endswith('()'):
                    p = (y_mt_0 & x_mt_0).sum() / y.size
                    top_right = f"{p:.0%} {top_right.strip('()')}"
                    replacement = '()'
                else:
                    replacement = None
                plt.text(xpos(xright), ypos(ytop), top_right, color=CABBI_colors.grey.shade(75).RGBn,
                     horizontalalignment='right', verticalalignment='top',
                     fontsize=10, fontweight='bold', zorder=10)
                top_right = replacement
            if ylb < 0. and xub > 0. and bottom_right:
                if bottom_right.endswith('()'):
                    p = (y_lt_0 & x_mt_0).sum() / y.size
                    bottom_right = f"{p:.0%} {bottom_right.strip('()')}"
                    replacement = '()'
                else:
                    replacement = None
                plt.text(xpos(xright), ypos(ybottom), bottom_right, color=colors.red.shade(50).RGBn,
                     horizontalalignment='right', verticalalignment='bottom',
                     fontsize=10, fontweight='bold', zorder=10)
                bottom_right = replacement
    plt.subplots_adjust(
        hspace=0, wspace=0,
        top=0.98, bottom=0.15,
        left=0.1, right=0.98,
    )
    if titles:
        plt.subplots_adjust(
            top=0.90,
        )
        for ax, letter in zip(axes[0, :], titles):
            plt.sca(ax)
            ylb, yub = plt.ylim()
            xlb, xub = plt.xlim()
            plt.text((xlb + xub) * 0.5, ylb + (yub - ylb) * 1.17, letter, color=letter_color,
                      horizontalalignment='center', verticalalignment='center',
                      fontsize=12, fontweight='bold')

def plot_feedstock_conventional_comparison_kde():
    plot_kde(
        'O1 - S1',
        yticks=[-20, -10, 0, 10, 20, 30, 40],
        xticks=[-0.40, -0.3, -0.20, -0.10, 0, 0.10, 0.20],
        top_left='Oilcane Favored',
        bottom_right='Sugarcane\nFavored',
        top_right='GWP\nTradeoff()',
        bottom_left='MFPP\nTradeoff()',
    )
    file = os.path.join(images_folder, 'feedstock_conventional_comparison_kde.png')
    plt.savefig(file, transparent=True)

def plot_feedstock_cellulosic_comparison_kde():
    plot_kde(
        'O2 - S2',
        yticks=[-40, -20, 0, 20, 40, 60, 80],
        xticks=[-5, -4, -3, -2, -1, 0],
        top_left='Oilcane Favored',
        bottom_right='Sugarcane Favored',
        top_right='GWP\nTradeoff()',
        bottom_left='MFPP\nTradeoff()',
    )
    file = os.path.join(images_folder, 'feedstock_cellulosic_comparison_kde.png')
    plt.savefig(file, transparent=True)

def plot_feedstock_comparison_kde():
    plot_kde_2d(
        ('O1 - S1', 'O2 - S2'),
        yticks=[[-20, -10, 0, 10, 20, 30, 40, 50]],
        xticks=[[-0.40, -0.3, -0.20, -0.10, 0, 0.10, 0.20],
                [-7.5, -6, -4.5, -3, -1.5, 0., 1.5, 3]],
        top_right='GWP\nTradeoff()',
        bottom_left='MFPP\nTradeoff()',
        top_left='Oilcane\nFavored()',
        bottom_right='\nSugarcane\nFavored()',
        titles=['(A) Conventional', '(B) Cellulosic'],
    )
    plt.subplots_adjust(
        wspace=0,
        
    )
    file = os.path.join(images_folder, 'feedstock_comparison_kde.png')
    plt.savefig(file, transparent=True)

def plot_configuration_comparison_kde():
    plot_kde(
        'O1 - O2',
        yticks=[-20, 0, 20, 40, 60],
        xticks=[-7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3],
        top_right='GWP\nTradeoff()',
        bottom_left='MFPP\nTradeoff()',
        top_left='Conventional\nFavored()',
        bottom_right='\nCellulosic\nFavored()',
    )
    file = os.path.join(images_folder, 'configuration_comparison_kde.png')
    plt.savefig(file, transparent=True)
    
def plot_crude_configuration_comparison_kde():
    plot_kde_2d(
        ('O1 - O3', 'O2 - O4'),
        yticks=[[-10, 0, 10, 20, 30, 40]],
        xticks=[
            [-2, -1.5, -1, -.5, 0],
            [-3, -2.5, -2, -1.5, -1, -.5, 0]
        ],
        top_right='GWP\nTradeoff()',
        bottom_left='MFPP\nTradeoff()',
        top_left='Biodiesel\nProduction Favored()',
        bottom_right='Crude Oil\nProduction Favored()',
        titles=['(A) Conventional', '(B) Cellulosic'],
    )
    file = os.path.join(images_folder, 'crude_configuration_comparison_kde.png')
    plt.savefig(file, transparent=True)

def plot_agile_comparison_kde():
    plot_kde_2d(
        ('O1* - O1', 'O2* - O2'),
        metrics=[TCI, MFPP],
        yticks=[[0, 3, 6, 9, 12, 15]],
        xticks=2*[[-150, -125, -100, -75, -50, -25, 0]],
        top_right='TCI-Tradeoff()',
        bottom_left='MFPP\nTradeoff()',
        top_left='Sorghum\nIntegration Favored()',
        bottom_right='Cane-only\nFavored()',
        ybox_kwargs=dict(light=CABBI_colors.green_dirty.RGBn, 
                         dark=CABBI_colors.green_dirty.shade(60).RGBn),
        titles=['(A) Conventional', '(B) Cellulosic'],
    )
    file = os.path.join(images_folder, 'agile_conventional_comparison_kde.png')
    plt.savefig(file, transparent=True)

def plot_open_comparison_kde(overlap=False):
    metrics = [MFPP, TCI, GWP_ethanol, biodiesel_production]
    df_conventional_oc = oc.get_monte_carlo('O1', metrics)
    df_cellulosic_oc = oc.get_monte_carlo('O2', metrics)
    df_conventional_sc = oc.get_monte_carlo('S1', metrics)
    df_cellulosic_sc = oc.get_monte_carlo('S2', metrics)
    MFPPi = MFPP.index
    TCIi = TCI.index
    if overlap:
        ys = np.zeros([1, 2], dtype=object)
        xs = np.zeros([1, 2], dtype=object)    
        ys[0, 0] = (df_conventional_oc[MFPPi], df_cellulosic_oc[MFPPi])
        ys[0, 1] = (df_conventional_sc[MFPPi], df_cellulosic_sc[MFPPi])
        xs[0, 0] = (df_conventional_oc[TCIi], df_cellulosic_oc[TCIi])
        xs[0, 1] = (df_conventional_sc[TCIi], df_cellulosic_sc[TCIi])
        yticks = [[-30, -15, 0, 15, 30, 45, 60, 75]]
        xticks = 2*[[200, 300, 400, 500, 600]]
    else:
        ys = np.array([
            [df_conventional_oc[MFPPi], df_conventional_sc[MFPPi]],
            [df_cellulosic_oc[MFPPi], df_cellulosic_sc[MFPPi]]
        ])
        xs = np.array([
            [df_conventional_oc[TCIi], df_conventional_sc[TCIi]],
            [df_cellulosic_oc[TCIi], df_cellulosic_sc[TCIi]]
        ])
        yticks = 2*[[-30, -15, 0, 15, 30, 45, 60, 75]]
        xticks = 2*[[200, 300, 400, 500, 600]]
    bst.plots.plot_kde_2d(
        ys=ys, xs=xs, xticks=xticks, yticks=yticks,
        xbox_kwargs=[dict(position=1), dict(position=1)],
        ybox_kwargs=[dict(position=0), dict(position=0)],
    )

#%%  General Monte Carlo box plots

def plot_monte_carlo_across_coordinate(coordinate, data, color_wheel):
    if isinstance(data, list):
        return [plot_monte_carlo_across_coordinate(coordinate, i, color_wheel) for i in data]
    else:
        color = color_wheel.next()
        return bst.plots.plot_montecarlo_across_coordinate(
            coordinate, data,
            light_color=color.tint(50).RGBn,
            dark_color=color.shade(50).RGBn,
        )

def monte_carlo_box_plot(data, positions, light_color, dark_color, width=None, 
                         hatch=None, outliers=False, **kwargs):
    if width is None: width = 0.8
    if outliers:
        flierprops = {'marker':'D',
                  'markerfacecolor': light_color,
                  'markeredgecolor': dark_color,
                  'markersize':3}
    else:
        flierprops = {'marker':''}
    bp = plt.boxplot(
        x=data, positions=positions, patch_artist=True,
        widths=width, whis=[5, 95],
        boxprops={'facecolor':light_color,
                  'edgecolor':dark_color},
        medianprops={'color':dark_color,
                     'linewidth':1.5},
        flierprops=flierprops,
        **kwargs
    )
    if hatch:
        for box in bp['boxes']:
            box.set(hatch = hatch)

def get_monte_carlo_key(index, dct, with_units=False):
    key = index[1] if with_units else index[1].split(' [')[0]
    if key in dct: key = f'{key}, {index[0]}'
    return key

def monte_carlo_results(with_units=False):
    results = {}    
    for name in oc.configuration_names + oc.comparison_names + oc.other_comparison_names + ('O3', 'O4', 'O1 - O3', 'O2 - O4'):
        try: 
            df = get_monte_carlo(name)
        except:
            warn(f'could not load {name}', RuntimeWarning)
            continue
        results[name] = dct = {}
        if name in ('O1', 'O2'):
            index = ethanol_over_biodiesel.index
            key = get_monte_carlo_key(index, dct, with_units)
            data = df[ethanol_production.index].values / df[biodiesel_production.index].values
            q05, q25, q50, q75, q95 = np.percentile(data, [5,25,50,75,95], axis=0)
            dct[key] = {
                'mean': np.mean(data),
                'std': np.std(data),
                'q05': q05,
                'q25': q25,
                'q50': q50,
                'q75': q75,
                'q95': q95,
            }
        for metric in (*tea_monte_carlo_metric_mockups, *tea_monte_carlo_derivative_metric_mockups,
                       *lca_monte_carlo_metric_mockups, *lca_monte_carlo_derivative_metric_mockups,
                       variables.GWP_ethanol_displacement, variables.GWP_ethanol_allocation):
            index = metric.index
            data = df[index].values
            q05, q25, q50, q75, q95 = np.percentile(data, [5,25,50,75,95], axis=0)
            key = get_monte_carlo_key(index, dct, with_units)
            dct[key] = {
                'mean': np.mean(data),
                'std': np.std(data),
                'q05': q05,
                'q25': q25,
                'q50': q50,
                'q75': q75,
                'q95': q95,
            }
    try:
        df_O2O1 = get_monte_carlo('O2 - O1')
        df_O1 = get_monte_carlo('O1')
    except:
        warn('could not load O2 - O1', RuntimeWarning)
    else:
        results['(O2 - O1) / O1'] = relative_results = {}
        for metric in (biodiesel_production, ethanol_production):
            index = metric.index
            key = index[1] if with_units else index[1].split(' [')[0]
            data = (df_O2O1[index].values / df_O1[index].values)
            q05, q25, q50, q75, q95 = np.percentile(data, [5,25,50,75,95], axis=0)
            relative_results[key] = {
                'mean': np.mean(data),
                'std': np.std(data),
                'q05': q05,
                'q25': q25,
                'q50': q50,
                'q75': q75,
                'q95': q95,
            }
    return results

def plot_monte_carlo(derivative=False, absolute=True, comparison=True,
                     configuration_names=None, comparison_names=None,
                     metrics=None, labels=None, tickmarks=None, agile=True, 
                     ncols=1, expand=None, step_min=None,
                     agile_only=False, xrot=None,
                     color_wheel=None, axes_box=None):
    if derivative:
        default_configuration_names = ['O1', 'O2']
        default_comparison_names = ['O2 - O1']
        metric_info = mc_derivative_metric_settings
        default_metrics = list(metric_info)
    else:
        default_configuration_names = oc.configuration_names[:-2]
        default_comparison_names = oc.comparison_names
        if comparison:
            metric_info = mc_comparison_settings
        else:
            metric_info = mc_metric_settings
        if agile_only:
            default_configuration_names = [i for i in default_configuration_names if '*' in i]
            default_comparison_names = [i for i in default_comparison_names if '*' in i]
            default_metrics = ['MFPP', 'TCI', 'production']
        else:
            default_metrics = list(metric_info)
    if configuration_names is None: configuration_names = default_configuration_names
    if comparison_names is None: comparison_names = default_comparison_names
    if metrics is None: metrics = default_metrics
    combined = absolute and comparison
    if agile_only:
        configuration_names = [i for i in configuration_names if '*' in i]
        comparison_names = [i for i in comparison_names if '*' in i]
    elif not agile:
        configuration_names = [i for i in configuration_names if '*' not in i]
        comparison_names = [i for i in comparison_names if '*' not in i]
    if combined:
        columns = configurations = configuration_names + comparison_names
    elif absolute:
        columns = configurations = configuration_names
    elif comparison:
        columns = configurations = comparison_names
    else:
        columns = configurations = []
    rows, ylabels, factors = zip(*[metric_info[i] for i in metrics])
    factors = [(i, j) for i, j in enumerate(factors) if j is not None]
    if color_wheel is None: color_wheel = CABBI_colors.wheel()
    N_rows = len(rows)
    if axes_box is None:
        fig, axes_box = plt.subplots(ncols=ncols, nrows=int(round(N_rows / ncols)))
        plt.subplots_adjust(wspace=0.45)
    else:
        fig = None
    axes = axes_box.transpose()
    axes = axes.flatten()
    N_cols = len(columns)    
    xtext = labels or [format_name(i).replace(' ', '') for i in configurations]
    N_marks = len(xtext)
    xticks = tuple(range(N_marks))
    
    def get_data(metric, name):
        try:
            df = get_monte_carlo(name, metric)
        except:
            return np.zeros([1, 1])
        else:
            values = df.values
            return values
    
    def plot(arr, position):
        if arr.ndim == 2:
            N = arr.shape[1]
            width = 0.618 / N
            boxwidth = 0.618 / (N + 1/N)
            plots = []
            for i in range(N):
                color = color_wheel.next()
                boxplot = monte_carlo_box_plot(
                    data=arr[:, i], positions=[position + (i-(N-1)/2)*width], 
                    light_color=color.RGBn, 
                    dark_color=color.shade(60).RGBn,
                    width=boxwidth,
                    hatch=getattr(color, 'hatch', None),
                )
                plots.append(boxplot)
            return plots
        else:
            color = color_wheel.next()
            return monte_carlo_box_plot(
                data=arr, positions=[position], 
                light_color=color.RGBn, 
                dark_color=color.shade(60).RGBn,
                width=0.618,
            )
    
    data = np.zeros([N_rows, N_cols], dtype=object)
    data[:] = [[get_data(i, j) for j in columns] for i in rows]
    for i, j in factors: data[i, :] *= j
        
    if tickmarks is None: 
        tickmarks = [
            bst.plots.rounded_tickmarks_from_data(
                i, step_min=step_min, N_ticks=8, lb_max=0, center=0,
                f=roundsigfigs, expand=expand,
                f_min=lambda x: np.percentile(x, 5),
                f_max=lambda x: np.percentile(x, 95),
            ) 
            for i in data
        ]

    x0 = len(configuration_names) - 0.5
    xf = len(columns) - 0.5
    for i in range(N_rows):
        ax = axes[i]
        plt.sca(ax)
        if combined:
            bst.plots.plot_vertical_line(x0)
            ax.axvspan(x0, xf, color=colors.purple_tint.tint(60).RGBn)
        plt.xlim(-0.5, xf)

    for j in range(N_cols):
        color_wheel.restart()
        for i in range(N_rows):
            ax = axes[i]
            plt.sca(ax)
            plot(data[i, j], j)
            plt.ylabel(ylabels[i])
    
    for i in range(N_rows):
        ax = axes[i]
        plt.sca(ax)
        yticks = tickmarks[i]
        plt.ylim([yticks[0], yticks[1]])
        if yticks[0] < 0.:
            bst.plots.plot_horizontal_line(0, color=CABBI_colors.black.RGBn, lw=0.8, linestyle='--')
        try:
            xticklabels = xtext if ax in axes_box[-1] else []
        except:
            xticklabels = xtext if i == N_rows - 1 else []
        bst.plots.style_axis(ax,  
            xticks = xticks,
            yticks = yticks,
            xticklabels= xticklabels, 
            ytick0=False,
            ytickf=False,
            offset_xticks=True,
            xrot=xrot,
        )
    if fig is None:
        fig = plt.gcf()
    else:
        plt.subplots_adjust(hspace=0)
    fig.align_ylabels(axes)
    return fig, axes

#%% Spearman

def plot_spearman(configurations, top=None, labels=None, metric=None, cutoff=None,
                  kind='TEA'):
    if metric is None:
        if kind == 'TEA':
            metric = MFPP
            metric_name = metric.name
        elif kind == 'LCA':
            metric = GWP_economic
            metric_name = r'GWP$_{\mathrm{economic}}$'
        else:
            raise ValueError(f"invalid kind '{kind}'")
    else:
        if metric == 'MFPP':
            metric = MFPP
        elif metric == 'GWP':
            metric = GWP_economic
        metric_name = metric.name
    stream_price = format_units('USD/gal')
    USD_ton = format_units('USD/ton')
    ng_price = format_units('USD/cf')
    electricity_price = format_units('USD/kWhr')
    operating_days = format_units('day/yr')
    capacity = format_units('dry-MMton/yr')
    titer = format_units('g/L')
    productivity = format_units('g/L/hr')
    material_GWP = '$\\mathrm{kg} \\cdot \\mathrm{CO}_{2}\\mathrm{eq} \\cdot \\mathrm{kg}^{-1}$'
    feedstock_GWP = '$\\mathrm{g} \\cdot \\mathrm{CO}_{2}\\mathrm{eq} \\cdot \\mathrm{kg}^{-1}$'
    index, ignored_list = zip(*[
         ('Bagasse oil retention [40 $-$ 70 %]', ['S2', 'S1', 'S2*', 'S1*']),
         ('Oil extraction efficiency [baseline + 0 $-$ 20 %]', ['S2', 'S1', 'S2*', 'S1*']),
        (f'Cane operating days [120 $-$ 180 {operating_days}]', []),
        (f'Sorghum operating days [30 $-$ 60 {operating_days}]', ['S2', 'S1', 'O1', 'O2']),
        (f'Crushing capacity [1.32 $-$ 2.20 {capacity}]', []),
        (f'Ethanol price [1.02, 1.80, 2.87 {stream_price}]', []),
        (f'Relative biodiesel price [0.31, 2.98, 4.11 {stream_price}]', []),
        (f'Natural gas price [3.71, 4.73, 6.18 {ng_price}]', ['S1', 'O1', 'S1*', 'O1*']),
        (f'Electricity price [0.0583, 0.065, 0.069 {electricity_price}]', ['S2', 'O2', 'S2*', 'O2*']),
         ('IRR [10 $-$ 15 %]', []),
        (f'Crude glycerol price [91 $-$ 200 {USD_ton}]', ['S2', 'S1', 'S2*', 'S1*']),
        (f'Pure glycerol price [501 $-$ 678 {USD_ton}]', ['S2', 'S1', 'S2*', 'S1*']),
         ('Saccharification reaction time [54 $-$ 90 hr]', ['S1', 'O1', 'S1*', 'O1*']),
        (f'Cellulase price [144 $-$ 240 {USD_ton}]', ['S1', 'O1', 'S1*', 'O1*']),
         ('Cellulase loading [1.5 $-$ 2.5 wt. % cellulose]', ['S1', 'O1', 'S1*', 'O1*']),
         ('PTRS base cost [14.9 $-$ 24.7 MMUSD]', ['S1', 'O1', 'S1*', 'O1*']),
         # ('Pretreatment reactor system base cost [14.9 $-$ 24.7 MMUSD]', ['S1', 'O1', 'S1*', 'O1*']),
         ('Cane glucose yield [85 $-$ 97.5 %]', ['S1', 'O1', 'S1*', 'O1*']),
         ('Sorghum glucose yield [85 $-$ 97.5 %]', ['S1', 'O1', 'S1*', 'O1*']),
         ('Cane xylose yield [65 $-$ 97.5 %]', ['S1', 'O1', 'S1*', 'O1*']),
         ('Sorghum xylose yield [65 $-$ 97.5 %]', ['S1', 'O1', 'S1*', 'O1*']),
         ('Glucose to ethanol yield [90 $-$ 95 %]', ['S1', 'O1', 'S1*', 'O1*']),
         ('Xylose to ethanol yield [50 $-$ 95 %]', ['S1', 'O1', 'S1*', 'O1*']),
        (f'Titer [65 $-$ 130 {titer}]', ['S1', 'O1', 'S1*', 'O1*']),
        (f'Productivity [1.0 $-$ 2.0 {productivity}]', ['S1', 'O1', 'S1*', 'O1*']),
         ('Cane PL content [7.5 $-$ 12.5 %]', ['S2', 'S1', 'S2*', 'S1*']),
         ('Sorghum PL content [7.5 $-$ 12.5 %]', ['S2', 'S1', 'S2*', 'S1*']),
         ('Cane FFA content [7.5 $-$ 12.5 %]', ['S2', 'S1', 'S2*', 'S1*']),
         ('Sorghum FFA content [7.5 $-$ 12.5 %]', ['S2', 'S1', 'S2*', 'S1*']),
         ('Cane oil content [5 $-$ 15 dry wt. %]', ['S2', 'S1', 'S2*', 'S1*']),
         ('Relative sorghum oil content [-3 $-$ 0 dry wt. %]', ['S2', 'S1', 'S2*', 'S1*', 'O2', 'O1']),
         ('TAG to FFA conversion [17.25 $-$ 28.75 % theoretical]', ['S1', 'O1', 'S1*', 'O1*']),
        # TODO: change lower upper values to baseline +- 10%
        (f'Feedstock GWPCF [26.3 $-$ 44.0 {feedstock_GWP}]', ['S1', 'S2', 'S1*', 'S2*']),
        (f'Methanol GWPCF [0.338 $-$ 0.563 {material_GWP}]', ['S1', 'S2', 'S1*', 'S2*']),
        (f'Pure glycerine GWPCF [1.25 $-$ 2.08 {material_GWP}]', ['S1', 'S2', 'S1*', 'S2*']),
        (f'Cellulase GWPCF [6.05 $-$ 10.1 {material_GWP}]', ['S1', 'O1', 'S1*', 'O1*']),
        (f'Natural gas GWPCF [0.297 $-$ 0.363 {material_GWP}]', ['S1', 'O1', 'S1*', 'O1*']),
    ])
    ignored_dct = {
        'S1': [],
        'O1': [],
        'S2': [],
        'O2': [],
        'S1*': [],
        'O1*': [],
        'S2*': [],
        'O2*': [],
    }
    for i, ignored in enumerate(ignored_list):
        for name in ignored: ignored_dct[name].append(i)
        index_name = index[i]
        if kind == 'LCA':
            for term in ('cost', 'price', 'IRR', 'time', 'capacity'):
                if term in index_name:
                    for name in ignored_dct: ignored_dct[name].append(i)
                    break
        elif kind == 'TEA':
            if 'GWP' in index_name:
                for name in ignored_dct: ignored_dct[name].append(i)
        else:
            raise ValueError(f"invalid kind '{kind}'")
    
    rhos = []
    for name in configurations:
        file = spearman_file(name)
        try: 
            df = pd.read_excel(file, header=[0, 1], index_col=[0, 1])
        except: 
            warning = RuntimeWarning(f"file '{file}' not found")
            warn(warning)
            continue
        s = df[metric.index]
        s.iloc[ignored_dct[name]] = 0.
        rhos.append(s)
    color_wheel = [CABBI_colors.orange, CABBI_colors.green_soft, CABBI_colors.blue, CABBI_colors.brown]
    fig, ax = bst.plots.plot_spearman_2d(rhos, top=top, index=index, cutoff=cutoff, 
                                         color_wheel=color_wheel,
                                         name=metric_name)
    plt.legend(
        handles=[
            mpatches.Patch(
                color=color_wheel[i].RGBn, 
                label=labels[i] if labels else format_name(configurations[i])
            )
            for i in range(len(configurations))
        ], 
        loc='lower left'
    )
    return fig, ax

# %% Other

def plot_configuration_breakdown(name, across_coordinate=False, **kwargs):
    oc.load(name)
    if across_coordinate:
        return bst.plots.plot_unit_groups_across_coordinate(
            oc.set_cane_oil_content,
            [5, 7.5, 10, 12.5],
            'Feedstock oil content [dry wt. %]',
            oc.unit_groups,
            colors=[area_colors[i.name].RGBn for i in oc.unit_groups],
            hatches=[area_hatches[i.name] for i in oc.unit_groups],
            **kwargs,
        )
    else:
        def format_total(x):
            if x < 1e3:
                return format(x, '.3g')
            else:
                x = int(x)
                n = 10 ** (len(str(x)) - 3)
                value = int(round(x / n) * n)
                return format(value, ',')
        for i in oc.unit_groups: 
            if i.name == 'EtOH prod.':
                i.name = 'Ethanol production'
            elif i.name == 'Oil ext.':
                i.name = 'Oil extraction'
            elif i.name == 'Biod. prod.':
                i.name = 'Biodiesel production'
            i.metrics[0].name = 'Inst. eq.\ncost'
            i.metrics[3].name = 'Elec.\ncons.'
            i.metrics[4].name = 'Mat.\ncost'
        
        return bst.plots.plot_unit_groups(
            oc.unit_groups,
            colors=[area_colors[i.name].RGBn for i in oc.unit_groups],
            hatches=[area_hatches[i.name] for i in oc.unit_groups],
            format_total=format_total,
            fraction=True,
            legend_kwargs=dict(
                loc='lower center',
                ncol=4,
                bbox_to_anchor=(0, -0.52),
                labelspacing=1.5, handlelength=2.8,
                handleheight=1, scale=0.8,
            ),
            **kwargs,
        )

def plot_TCI_areas_across_oil_content(configuration='O2'):
    oc.load(configuration)
    data = {i.name: [] for i in oc.unit_groups}
    increasing_areas = []
    decreasing_areas = []
    oil_contents = np.linspace(5, 15, 10)
    for i in oil_contents:
        oc.set_cane_oil_content(i)
        oc.sys.simulate()
        for i in oc.unit_groups: data[i.name].append(i.get_installed_cost())
    for name, group_data in data.items():
        lb, *_, ub = group_data
        if ub > lb: 
            increasing_areas.append(group_data)
        else:
            decreasing_areas.append(group_data)
    increasing_values = np.sum(increasing_areas, axis=0)
    increasing_values -= increasing_values[0]
    decreasing_values = np.sum(decreasing_areas, axis=0)
    decreasing_values -= decreasing_values[-1]
    plt.plot(oil_contents, increasing_values, label='Oil & fiber areas')
    plt.plot(oil_contents, decreasing_values, label='Sugar areas')
    
# def plot_monte_carlo_across_oil_content(kind=0, derivative=False):
#     MFPP, TCI, *production, electricity_production, natural_gas_consumption = tea_monte_carlo_metric_mockups
#     rows = [MFPP, TCI, production]
#     if kind == 0:
#         columns = across_oil_content_names
#     elif kind == 1:
#         columns = across_oil_content_agile_names
#     elif kind == 2:
#         columns = across_oil_content_comparison_names
#     elif kind == 3:
#         columns = across_oil_content_agile_comparison_names
#     elif kind == 4:
#         columns = across_oil_content_agile_direct_comparison_names
#     else:
#         raise NotImplementedError(str(kind))
#     if derivative:
#         x = 100 * (oil_content[:-1] + np.diff(oil_content) / 2.)
#         ylabels = [
#             f"MFPP der. [{format_units('USD/ton')}]",
#             f"TCI der. [{format_units('10^6*USD')}]",
#             f"Production der. [{format_units('gal/ton')}]"
#         ]
#     else:
#         x = 100 * oil_content
#         ylabels = [
#             f"MFPP$\backprime$ [{format_units('USD/ton')}]",
#             f"TCI [{format_units('10^6*USD')}]",
#             f"Production [{format_units('gal/ton')}]"
#         ]
#     N_cols = len(columns)
#     N_rows = len(rows)
#     fig, axes = plt.subplots(ncols=N_cols, nrows=N_rows)
#     data = np.zeros([N_rows, N_cols], dtype=object)
    
#     def get_data(metric, name):
#         if isinstance(metric, bst.Variable):
#             return get_monte_carlo_across_oil_content(name, metric, derivative)
#         else:
#             return [get_data(i, name) for i in metric]
    
#     data = np.array([[get_data(i, j) for j in columns] for i in rows])
#     tickmarks = [None] * N_rows
#     get_max = lambda x: max([i.max() for i in x]) if isinstance(x, list) else x.max()
#     get_min = lambda x: min([i.min() for i in x]) if isinstance(x, list) else x.min()
#     N_ticks = 5
#     for r in range(N_rows):
#         lb = min(min([get_min(i) for i in data[r, :]]), 0)
#         ub = max([get_max(i) for i in data[r, :]])
#         diff = 0.1 * (ub - lb)
#         ub += diff
#         if derivative:
#             lb = floor(lb)
#             ub = ceil(ub)
#             step = (ub - lb) / (N_ticks - 1)
#             tickmarks[r] = [0, 1] if step == 0 else [int(lb + step * i) for i in range(N_ticks)]
#         else:
#             if rows[r] is MFPP:
#                 if kind == 0 or kind == 1:
#                     tickmarks[r] = [-20, 0, 20, 40, 60]
#                 elif kind == 2:
#                     tickmarks[r] = [-20, -10, 0, 10, 20]
#                 elif kind == 3:
#                     tickmarks[r] = [-10, 0, 10, 20, 30]
#                 elif kind == 4:
#                     tickmarks[r] = [-5, 0, 5, 10, 15]
#                 continue
#             lb = floor(lb / 15) * 15
#             ub = ceil(ub / 15) * 15
#             step = (ub - lb) / (N_ticks - 1)
#             tickmarks[r] = [0, 1] if step == 0 else [int(lb + step * i) for i in range(N_ticks)]
#     color_wheel = CABBI_colors.wheel()
#     for j in range(N_cols):
#         color_wheel.restart()
#         for i in range(N_rows):
#             arr = data[i, j]
#             ax = axes[i, j]
#             plt.sca(ax)
#             percentiles = plot_monte_carlo_across_coordinate(x, arr, color_wheel)
#             if i == 0: ax.set_title(format_name(columns[j]))
#             xticklabels = i == N_rows - 1
#             yticklabels = j == 0
#             if xticklabels: plt.xlabel('Oil content [dry wt. %]')
#             if yticklabels: plt.ylabel(ylabels[i])
#             bst.plots.style_axis(ax,  
#                                  xticks = [5, 10, 15],
#                                  yticks = tickmarks[i],
#                                  xticklabels= xticklabels, 
#                                  yticklabels= yticklabels,
#                                  ytick0=False)
#     for i in range(N_cols): fig.align_ylabels(axes[:, i])
#     plt.subplots_adjust(hspace=0.1, wspace=0.1)


# -*- coding: utf-8 -*-
"""
Created on Thu Jun 27 23:12:04 2019

@author: yoelr
"""
import thermosteam as tmo
from biorefineries import lipidcane as lc 
from thermosteam import functional as fn
import pandas as pd

__all__ = ('create_chemicals', 'get_grouped_chemicals', 'chemical_groups')

# %% Constants

# Common structural carbohydrates properties

# Assume heat capacity of lignin, cellulose, and hemicellulose
# and all components at 350 K are about the same.
Cp_cellulosic = 1.364

# Assume density is similar for most solids
rho = 1540 # kg/m3

cal2joule = 4.184

chemical_groups = dict(
        OtherSugars = ('Arabinose',
                       'Mannose',
                       'Galactose',
                       'Cellobiose',
                       'Sucrose'),
        SugarOligomers = ('GlucoseOligomer',
                          'XyloseOligomer',
                          'GalactoseOligomer',
                          'ArabinoseOligomer',
                          'MannoseOligomer'),
        OrganicSolubleSolids = ('AmmoniumAcetate',
                                'SolubleLignin',
                                'Extract', 
                                'LacticAcid', 
                                'Cellulase'),
        InorganicSolubleSolids = ('AmmoniumSulfate',
                                  'DAP',
                                  'NaOH',
                                  'HNO3',
                                  'NaNO3'),
        Furfurals = ('Furfural',
                     'HMF'),
        OtherOrganics = ('Glycerol',
                         'Denaturant',
                         'Oil',
                         'SuccinicAcid',
                         'Xylitol'),
        COxSOxNOxH2S = ('NO',
                        'NO2',
                        'SO2',
                        'CO',
                        'H2S'),
        Protein = ('Protein',
                   'Enzyme',
                   'DenaturedEnzyme'),
        CellMass = ('WWTsludge',
                    'Z_mobilis',
                    'T_reesei'),
        OtherInsolubleSolids = ('Tar',
                                'Ash',
                                'Lime'),
        OtherStructuralCarbohydrates = ('Arabinan', 
                                        'Mannan', 
                                        'Galactan')
)

def get_grouped_chemicals(stream, units='kmol/hr'):
    new_stream = tmo.Stream(stream.thermo)
    new_stream.set_flow(stream.mol, units, stream.chemicals.IDs)
    data = {group: new_stream.get_flow(units, IDs).sum() for group, IDs in chemical_groups.items()}
    return pd.Series(data)


# %% Chemicals object and define functions

def create_chemicals():
    chems = tmo.Chemicals([])
    
    def append_single_phase_chemical(ID, search_ID=None, **data):
        chemical = tmo.Chemical(ID, search_ID=search_ID, **data)
        try: chemical.at_state(phase=chemical.phase_ref)
        except: pass
        chemical.default()    
        chems.append(chemical)
    
    def extend_single_phase_chemicals(IDs, **data):
        for ID in IDs: append_single_phase_chemical(ID, **data)
    
    def append_new_single_phase_chemical(ID, source=None, **data):
        chemical = tmo.Chemical.blank(ID, **data)
        if source: 
            default_phase_ref = source.phase_ref
            chemical.copy_models_from(source)
        else:
            default_phase_ref = 'l'
        if not chemical.phase_ref:
            chemical.phase_ref = default_phase_ref
        chemical.at_state(chemical.phase_ref)
        chemical.default()
        chems.append(chemical)
    
    def append_chemical_copy(ID, chemical):
        new_chemical = chemical.copy(ID)
        chems.append(new_chemical)
    
    def set_Cp(single_phase_chemical, Cp):
        chem = single_phase_chemical
        chem.Cn.add_model(Cp * chem.MW, top_priority=True)
    
    def set_rho(single_phase_chemical, rho):
        V = fn.rho_to_V(rho, single_phase_chemical.MW)
        single_phase_chemical.V.add_model(V, top_priority=True)
    
    ### Define species
    
    # As is in data bank
    chems.extend(
        tmo.Chemicals([lc.chemicals.Water, lc.chemicals.Ethanol, 'AceticAcid',
                       'Furfural', lc.chemicals.Glycerol, 'H2SO4', 'NH3', 
                       'LacticAcid', 'SuccinicAcid', lc.chemicals.P4O10])
    )
    chems.H2SO4.at_state('l')
    append_single_phase_chemical('Lime', 'Ca(OH)2')
    append_single_phase_chemical('HNO3', 'NitricAcid')
    append_single_phase_chemical('NH4OH')
    append_single_phase_chemical('Denaturant', 'Octane')
    append_single_phase_chemical('DAP', 'Diammonium Phosphate')
    append_single_phase_chemical('AmmoniumAcetate')
    append_single_phase_chemical('AmmoniumSulfate')
    append_single_phase_chemical('NaNO3', 'SodiumNitrate')
    append_single_phase_chemical('Oil', 'Oleic acid')
    append_single_phase_chemical('HMF', 'Hydroxymethylfurfural', phase='l')
    
    # Will remain in the vapor phase
    extend_single_phase_chemicals(['N2', 'O2', 'CH4', 'H2S', 'SO2'])
    append_single_phase_chemical('CO2')
    
    # Analagous vapors
    append_new_single_phase_chemical('NO2', chems.N2,
                                     formula='NO2',
                                     Hf=7925*cal2joule)
    append_new_single_phase_chemical('NO', chems.N2,
                                     formula='NO',
                                     Hf=82.05)
    append_single_phase_chemical('CO', 'Carbon monoxide', Hf=-110.522)
    
    # Will remain as  solid
    extend_single_phase_chemicals(['Glucose', 'Xylose', 'Sucrose'], Hfus=0)
    append_single_phase_chemical('CaSO4')
    
    subgroup = chems['Glucose', 'Xylose', 'Sucrose', 'CaSO4', 'AmmoniumSulfate']
    for chemical in subgroup: set_Cp(chemical, Cp_cellulosic)
    
    # Analagous sugars
    append_chemical_copy('Mannose', chems.Glucose)
    append_chemical_copy('Galactose', chems.Glucose)
    append_chemical_copy('Arabinose', chems.Xylose)
    
    # Other analogues
    append_chemical_copy('CellulaseNutrients', chems.Glucose)
    append_chemical_copy('Extract', chems.Glucose)
    append_chemical_copy('Acetate', chems.AceticAcid)
    append_chemical_copy('Tar', chems.Xylose)
    chems.Acetate.Hf = -103373
    
    # Chemicals taken from previous study
    chems.append(lc.chemicals.Ash)
    chems.append(lc.chemicals.NaOH)
    append_new_single_phase_chemical('Lignin',
                                     formula='C8H8O3',
                                     Hf=-108248*cal2joule)
    set_rho(chems.Lignin, 1540)
    set_Cp(chems.Lignin, Cp_cellulosic)
    append_chemical_copy('SolubleLignin', chems.Lignin)
    
    # Create structural carbohydrates
    append_chemical_copy('GlucoseOligomer', chems.Glucose)
    set_Cp(chems.GlucoseOligomer, Cp_cellulosic)
    chems.GlucoseOligomer._formula = None
    chems.GlucoseOligomer.formula = "C6H10O5"
    chems.GlucoseOligomer.Hf = -233200*cal2joule
    
    append_chemical_copy('GalactoseOligomer', chems.GlucoseOligomer)
    append_chemical_copy('MannoseOligomer', chems.GlucoseOligomer)
    append_chemical_copy('XyloseOligomer', chems.Xylose)
    set_Cp(chems.XyloseOligomer, Cp_cellulosic)
    chems.XyloseOligomer._formula =None
    chems.XyloseOligomer.formula = "C5H8O4"
    chems.XyloseOligomer.Hf = -182100*cal2joule
    
    append_chemical_copy('ArabinoseOligomer', chems.XyloseOligomer)
    
    # Other
    append_new_single_phase_chemical('Z_mobilis', formula="CH1.8O0.5N0.2",
                                     Hf=-31169.39*cal2joule)
    append_new_single_phase_chemical('T_reesei', formula="CH1.645O0.445N0.205S0.005",
                                     Hf=-23200.01*cal2joule)
    append_new_single_phase_chemical('Biomass', formula="CH1.64O0.39N0.23S0.0035",
                                     Hf=-23200.01*cal2joule)
    append_new_single_phase_chemical('Cellulose', formula="C6H10O5", # Glucose monomer minus water
                                     Hf=-233200.06*cal2joule)
    append_new_single_phase_chemical('Protein', formula="CH1.57O0.31N0.29S0.007",
                                     Hf=-17618*cal2joule)
    append_new_single_phase_chemical('Enzyme', formula="CH1.59O0.42N0.24S0.01",
                                     Hf=-17618*cal2joule)
    append_new_single_phase_chemical('Glucan', formula='C6H10O5',
                                     Hf=-233200*cal2joule)
    append_new_single_phase_chemical('Xylan', formula="C5H8O4",
                                     Hf=-182100*cal2joule)
    append_new_single_phase_chemical('Xylitol', formula="C5H12O5",
                                     Hf=-243145*cal2joule)
    append_new_single_phase_chemical('Cellobiose', formula="C12H22O11",
                                     Hf=-480900*cal2joule)
    append_new_single_phase_chemical('CSL', 
                                     formula='H2.8925O1.3275C1N0.0725S0.00175',
                                     Hf=(chems.Protein.Hf/4
                                         + chems.Water.Hf/2
                                         + chems.LacticAcid.Hf/4))
    append_chemical_copy('DenaturedEnzyme', chems.Enzyme)
    append_chemical_copy('Arabinan', chems.Xylan)
    append_chemical_copy('Mannan',   chems.Glucan)
    append_chemical_copy('Galactan', chems.Glucan)
    
    # TODO: Maybe remove this
    # For waste water
    append_chemical_copy('WWTsludge', chems.Biomass)
    append_chemical_copy('Cellulase', chems.Enzyme)
    
    # New feature in Thermosteam allows salt and solutes to be accounted 
    # for in VLE; leading to more accurate results. However, it is not included
    # here as there is no significant difference in results.
    # for i in chems['Glucose', 'Xylose']:
    #     i.N_solutes = 1
    # for i in chems['Sucrose', 'CaSO4', 'AmmoniumSulfate']:
    #     i.N_solutes = 2
        
    chems.compile()
    chems.set_synonym('Lime', 'Ca(OH)2')
    chems.set_synonym('Water', 'H2O')
    chems.set_synonym('H2SO4', 'SulfuricAcid')
    chems.set_synonym('NH3', 'Ammonia')
    chems.set_synonym('AmmoniumSulfate', '(NH4)2SO4')
    chems.set_synonym('Denaturant', 'Octane')
    chems.set_synonym('CO2', 'CarbonDioxide')
    return chems



# -*- coding: utf-8 -*-
"""
Created on Mon Feb  4 10:02:05 2019

@author: yoelr
"""
import biosteam as bst

__all__ = ('load_process_settings',)

# %% Process settings

def load_process_settings():
    bst.process_tools.default_utilities()
    bst.CE = 607.5 # 2019
    bst.PowerUtility.price = 0.065
    HeatUtility = bst.HeatUtility
    steam_utility = HeatUtility.get_agent('low_pressure_steam')
    steam_utility.heat_transfer_efficiency = 0.9
    steam_utility.regeneration_price = 0.30626
    steam_utility.T = 529.2
    steam_utility.P = 44e5
    HeatUtility.get_agent('cooling_water').regeneration_price = 0
    HeatUtility.get_agent('chilled_water').heat_transfer_price = 0

ton2kg = 1./907.18474 # ton/hr to kg/hr
ethanol_density_kgL = 0.789 # kg/L
liter_per_gallon = 3.78541
ethanol_cost = 2.15 # USD/gal
ethanol_density_kggal = liter_per_gallon * ethanol_density_kgL # kg/gal
enzyme_price = 4.24 * 50/1000 # USD/kg



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from . import (
    units,
    _process_settings,
    _chemicals,
    _system,
    _tea,
    _oil_extraction_specification,
    _distributions,
    _evaluation,
    _uncertainty_plots,
    _contour_plots,
    _lca_characterization_factors,
    _load_data,
    _parse_configuration,
)
from .units import *
from ._process_settings import *
from ._chemicals import *
from ._contour_plots import *
from ._system import *
from ._tea import *
from ._oil_extraction_specification import *
from ._distributions import *
from ._evaluation import *
from ._uncertainty_plots import *
from ._contour_plots import *
from ._lca_characterization_factors import *
from ._load_data import *
from ._parse_configuration import *
from ._variable_mockups import *

__all__ = (
    units.__all__,
    *_process_settings.__all__,
    *_chemicals.__all__,
    *_system.__all__,
    *_tea.__all__,
    *_oil_extraction_specification.__all__,
    *_distributions.__all__,
    *_evaluation.__all__,
    *_uncertainty_plots.__all__,
    *_contour_plots.__all__,
    *_lca_characterization_factors.__all__,
    *_load_data.__all__,
    *_parse_configuration.__all__,
    'sys',
    'tea', 
    'flowsheet',
)

import biosteam as bst
from biosteam.utils import MockStream
import thermosteam as tmo
from biorefineries.sugarcane import create_sugarcane_to_ethanol_system
from biorefineries.lipidcane import (
    set_lipid_fraction as set_oil_fraction, 
)
from biosteam import main_flowsheet, UnitGroup
from chaospy import distributions as shape
import numpy as np
from ._process_settings import load_process_settings
from ._chemicals import create_chemicals
from ._system import (
    create_oilcane_to_biodiesel_and_ethanol_1g,
    create_oilcane_to_biodiesel_and_ethanol_combined_1_and_2g_post_fermentation_oil_separation,
    create_sugarcane_to_ethanol_combined_1_and_2g,    
)
from ._parse_configuration import (
    parse,
    format_configuration,
)
from ._tea import (
    create_tea,
)
from ._oil_extraction_specification import (
    OilExtractionSpecification,
    MockExtractionSpecification,
)
from ._distributions import (
    ethanol_price_distribution,
    biodiesel_minus_ethanol_price_distribution,
    natural_gas_price_distribution,
    mean_biodiesel_price,
    mean_ethanol_price,
    mean_glycerol_price,
    mean_natural_gas_price,
    mean_electricity_price,
)
from ._lca_characterization_factors import (
    GWP_characterization_factors,
    set_GWPCF,
    GWP,
)
from ._tables import (
    save_detailed_expenditure_tables, 
    save_detailed_life_cycle_tables
)
from biorefineries import cornstover as cs

_system_loaded = False
_chemicals_loaded = False

PRS = cs.PretreatmentReactorSystem
PRS_cost_item = PRS.cost_items['Pretreatment reactor system']
kg_per_ton = 907.18474


configuration_names = (
    'S1', 'O1', 'S2', 'O2', 'S1*', 'O1*', 'S2*', 'O2*', 'O3', 'O4',
)
comparison_names = (
    # 'I - ∅', 
    'O1 - S1', 
    'O2 - S2', 
    'O2 - O1', 
    'O1* - O1', 
    'O2* - O2',  
)

other_comparison_names = (
    'O1* - S1*', 'O2* - S2*', 
)

across_oil_content_names = (
    'O1', 'O2', 
)

across_oil_content_agile_names = (
    'O1*', 'O2*', 
)

across_oil_content_comparison_names = (
    'O1 - S1', 'O2 - S2', 'O2 - O1', 
)

across_oil_content_agile_direct_comparison_names = (
    'O1* - O1', 'O2* - O2', 
)

across_oil_content_agile_comparison_names = (
    'O1* - S1*', 'O2* - S2*', 'O2* - O1*', 
)

def load_chemicals():
    global chemicals, _chemicals_loaded
    chemicals = create_chemicals()
    _chemicals_loaded = True

def disable_derivative(disable=True):
    global _derivative_disabled
    _derivative_disabled = disable
    
def enable_derivative(enable=True):
    global _derivative_disabled
    _derivative_disabled = not enable
    
_derivative_disabled = False

def load(name, cache={}, reduce_chemicals=True, 
         enhanced_cellulosic_performance=False,
         enhanced_biodiesel_production=False):
    dct = globals()
    number, agile = dct['configuration'] = configuration = parse(name)
    key = (number, agile, enhanced_cellulosic_performance)
    if cache is not None and key in cache:
        dct.update(cache[key])
        return
    global oilcane_sys, sys, tea, specs, flowsheet, _system_loaded
    global oil_extraction_specification, model, unit_groups
    global HXN, BT
    if not _chemicals_loaded: load_chemicals()
    flowsheet_name = format_configuration(configuration)
    if enhanced_cellulosic_performance:
        flowsheet_name += '_enhanced_fermentation'
    flowsheet = bst.Flowsheet(flowsheet_name)
    main_flowsheet.set_flowsheet(flowsheet)
    bst.settings.set_thermo(chemicals)
    load_process_settings()
    s = flowsheet.stream
    u = flowsheet.unit
    operating_hours = 24 * 180
    
    ## System
    
    area_names = None
    def rename_storage_units(storage):
        bst.rename_units([i for i in oilcane_sys.units if bst.is_storage_unit(i)], storage)
    
    if number == -1:
        isplit_efficiency_is_reversed = None
        # starting_chemicals = create_starting_chemicals()
        # bst.settings.set_thermo(starting_chemicals)
        oilcane_sys = create_sugarcane_to_ethanol_system(
            operating_hours=operating_hours,
            use_area_convention=True,
            pellet_bagasse=True,
        )
        area_names = [
            'Feedstock handling', 
            'Juicing', 
            'EtOH prod.', 
            'CH&P',
            'Utilities',
            'HXN',
            'Storage',
        ]
        rename_storage_units(700)
    elif number == -2:
        isplit_efficiency_is_reversed = None
        oilcane_sys = create_sugarcane_to_ethanol_combined_1_and_2g(
            operating_hours=operating_hours,
        )
        area_names = [
            'Feedstock handling', 
            'Juicing', 
            'Pretreatment',
            'EtOH prod.',
            'Wastewater treatment',
            'CH&P', 
            'Utilities',
            'HXN',
            'Storage',
        ]
        rename_storage_units(900)
    elif number == 1:
        isplit_efficiency_is_reversed = False
        oilcane_sys = create_oilcane_to_biodiesel_and_ethanol_1g(
            operating_hours=operating_hours,
        )
        area_names = [
            'Feedstock handling', 
            'Juicing', 
            'EtOH prod.', 
            'Oil ext.',
            'Biod. prod.', 
            'CH&P',
            'Utilities',
            'HXN',
            'Storage',
        ]
        rename_storage_units(1000)
    elif number == 2:
        isplit_efficiency_is_reversed = True
        area_names = [
            'Feedstock handling', 
            'Juicing', 
            'Pretreatment',
            'EtOH prod.',
            'Wastewater treatment',
            'Oil ext.',
            'CH&P', 
            'Biod. prod.',
            'Utilities',
            'HXN',
            'Storage',
        ]
        oilcane_sys = create_oilcane_to_biodiesel_and_ethanol_combined_1_and_2g_post_fermentation_oil_separation(
            operating_hours=operating_hours,
        )
        rename_storage_units(1100)
    elif number == 3:
        isplit_efficiency_is_reversed = False
        oilcane_sys = create_oilcane_to_crude_oil_and_ethanol_1g(
            operating_hours=operating_hours,
        )
        area_names = [
            'Feedstock handling', 
            'Juicing', 
            'EtOH prod.', 
            'Oil ext.', 
            'CH&P',
            'Utilities',
            'HXN',
            'Storage',
        ]
        rename_storage_units(800)
    elif number == 4:
        isplit_efficiency_is_reversed = True
        area_names = [
            'Feedstock handling', 
            'Juicing', 
            'Pretreatment',
            'EtOH prod.',
            'Wastewater treatment',
            'Oil ext.',
            'CH&P', 
            'Utilities',
            'HXN',
            'Storage',
        ]
        oilcane_sys = create_oilcane_to_crude_oil_and_ethanol_combined_1_and_2g_post_fermentation_oil_separation(
            operating_hours=operating_hours,
        )
        rename_storage_units(1000)
    else:
        raise NotImplementedError(number)
    
    oilcane_sys.set_tolerance(rmol=1e-5, mol=1e-3, subsystems=True)
    dct.update(flowsheet.to_dict())
    
    def get_stream(ID):
        if ID in flowsheet.stream:
            return s[ID]
        else:
            return MockStream(ID)
    
    crude_glycerol = get_stream('crude_glycerol')
    pure_glycerine = get_stream('pure_glycerine')
    if number == 1:
        oilcane_sys.prioritize_unit(u.T608)
    elif number == 2:
        oilcane_sys.prioritize_unit(u.T808)
    if number < 0:
        dct['feedstock'] = feedstock = s.sugarcane
        dct['oilcane_sys'] = oilcane_sys
    else:
        dct['feedstock'] = feedstock = s.oilcane
    unit_groups = UnitGroup.group_by_area(oilcane_sys.units)
    if area_names:
        for i, j in zip(unit_groups, area_names): i.name = j
    for i in unit_groups: i.autofill_metrics(shorthand=True)
    
    for BT in oilcane_sys.units:
        if isinstance(BT, bst.BoilerTurbogenerator): break

    HXN = None
    for HXN_group in unit_groups:
        if HXN_group.name == 'HXN':
            HXN_group.filter_savings = False
            HXN = HXN_group.units[0]
            assert isinstance(HXN, bst.HeatExchangerNetwork)
    unit_groups[-1].metrics[-1].getter = lambda: 0.    
    
    
    if abs(number) in (2, 4):
        prs, = flowsheet(cs.units.PretreatmentReactorSystem)
        saccharification, = flowsheet(cs.units.Saccharification)
        seed_train, = flowsheet(cs.units.SeedTrain)
        fermentor, = flowsheet(cs.units.CoFermentation)
        dct['pretreatment_rxnsys'] = tmo.ReactionSystem(
            prs.reactions, saccharification.saccharification
        )
        dct['fermentation_rxnsys'] = tmo.ReactionSystem(
            seed_train.reactions, fermentor.cofermentation
        )
        dct['cellulosic_rxnsys'] = tmo.ReactionSystem(
            prs.reactions, saccharification.saccharification,
            seed_train.reactions, fermentor.cofermentation
        )
        saccharification.saccharification.X[0] = 0.0 # Baseline
        prs.reactions.X[10] = 0.0 # baseline
            
    def set_glucose_yield(glucose_yield):
        if abs(number) in (2, 4):
            glucose_yield *= 0.01
            X1 = prs.reactions.X[0]
            X1_side = prs.reactions.X[1:3].sum()
            X2_side = saccharification.saccharification.X[:2].sum()
            saccharification.saccharification.X[2] = X2 = (glucose_yield - X1) / (1 - X1_side)
            X_excess = (X2_side + X2) - 1
            if X_excess > 0: breakpoint()
            
    def set_xylose_yield(xylose_yield):
        if abs(number) in (2, 4):
            xylose_yield *= 0.01
            X1_side = prs.reactions.X[9:11].sum()
            prs.reactions.X[8] = X1 = xylose_yield
            X_excess = (X1_side + X1) - 1
            if X_excess > 0.: breakpoint()
    
    if agile: 
        dct['oilsorghum'] = oilsorghum = bst.Stream(
            ID='oilsorghum', phase='l', T=298.15, P=101325, 
            Water=2.333e+05, Glucose=3703, Sucrose=4.196e+04, Ash=2000, 
            Cellulose=2.227e+04, Hemicellulose=1.314e+04, Lignin=1.193e+04, 
            Solids=5000, units='kg/hr'
        )
        
        sys = bst.AgileSystem()
        @sys.operation_parameter(mode_dependent=True)
        def set_oil_content(oil_content, mode):
            if number > 0: 
                set_oil_fraction(oil_content, feedstock,
                                 FFA_fraction=mode.FFA_content,
                                 z_mass_carbs_baseline=mode.z_mass_carbs_baseline,
                                 PL_fraction=mode.PL_content)
            else:
                F_mass = feedstock.F_mass
                feedstock.copy_flow(mode.feedstock)
                feedstock.F_mass = F_mass
        sys.operation_parameter(set_glucose_yield)
        sys.operation_parameter(set_xylose_yield)
        
        dct['cane_mode'] = cane_mode = sys.operation_mode(oilcane_sys,
            operating_hours=180*24, oil_content=0.05, feedstock=feedstock.copy(),
            z_mass_carbs_baseline=0.1491, glucose_yield=85, xylose_yield=65, 
            FFA_content=0.10, PL_content=0.10
        )
        dct['sorghum_mode'] = sorghum_mode = sys.operation_mode(oilcane_sys, 
            operating_hours=60*24, oil_content=0.05, glucose_yield=79, xylose_yield=86,
            feedstock=oilsorghum,
            z_mass_carbs_baseline=0.1371, FFA_content=0.10, PL_content=0.10,
        )
        tea = create_tea(sys)
        tea.operating_days = 260 
        tea.IRR = 0.10
    else:
        sys = oilcane_sys
        tea = create_tea(sys)
        tea.operating_days = 200
        tea.IRR = 0.10
        
        # def electricity_GWP():
        #     power_utility = bst.PowerUtility.sum([i.power_utility for i in sys.cost_units])
        #     return power_utility.get_impact(production_key=GWP) * sys.operating_hours
        
    tea.income_tax = 0.21 # Davis et al. 2018; https://www.nrel.gov/docs/fy19osti/71949.pdf
    
    ## Specification for analysis
    if number < 0:
        dct['biodiesel'] = bst.Stream('biodiesel')
        isplit_a = None
        isplit_b = None
        oil_extraction_specification = MockExtractionSpecification()
    else:
        if number > 2:
            dct['biodiesel'] = bst.Stream('biodiesel')
        
        for i in oilcane_sys.cost_units:
            if getattr(i, 'tag', None) == 'oil extraction efficiency':
                isplit_a = i.isplit
                break
        
        for i in oilcane_sys.cost_units:
            if getattr(i, 'tag', None) == 'bagasse oil retention':
                isplit_b = i.isplit
                break
        
        oil_extraction_specification = OilExtractionSpecification(
            sys, [feedstock], isplit_a, isplit_b, isplit_efficiency_is_reversed
        )
    
    ## LCA
    
    # Set non-negligible characterization factors
    if abs(number) not in (2, 4):
        for i in ('FGD_lime', 'cellulase', 'DAP', 'CSL', 'caustic'): MockStream(i)
    if number < 0 or number > 2:
        for i in ('catalyst', 'methanol', 'HCl', 'NaOH', 'crude_glycerol', 'pure_glycerine'): MockStream(i)
    if abs(number) not in (1, 3):
        MockStream('dryer_natural_gas')
        
    set_GWPCF(feedstock, 'sugarcane')
    set_GWPCF(s.H3PO4, 'H3PO4')
    set_GWPCF(s.lime, 'lime', dilution=0.046) # Diluted with water
    set_GWPCF(s.denaturant, 'gasoline')
    set_GWPCF(s.FGD_lime, 'lime', dilution=0.451)
    set_GWPCF(s.cellulase, 'cellulase', dilution=0.02) 
    set_GWPCF(s.DAP, 'DAP')
    set_GWPCF(s.CSL, 'CSL')
    set_GWPCF(s.caustic, 'NaOH', 0.5)
    set_GWPCF(s.catalyst, 'NaOH', 0.5)
    set_GWPCF(s.catalyst, 'methanol catalyst mixture')
    set_GWPCF(s.methanol, 'methanol')
    set_GWPCF(s.HCl, 'HCl')
    set_GWPCF(s.NaOH, 'NaOH')
    set_GWPCF(s.pure_glycerine, 'pure-glycerol')
    set_GWPCF(s.dryer_natural_gas, 'CH4')
    set_GWPCF(s.crude_glycerol, 'crude-glycerol', dilution=0.80)
    set_GWPCF(s.biodiesel, 'biodiesel displacement')
    bst.PowerUtility.set_CF(GWP, GWP_characterization_factors['Electricity'])
    dct['natural_gas_streams'] = natural_gas_streams = [s.natural_gas]
    if abs(number) in (1, 3): natural_gas_streams.append(s.dryer_natural_gas)
    for stream in natural_gas_streams:
        set_GWPCF(stream, 'CH4')
    
    ## Model
    model = bst.Model(sys, exception_hook='raise', retry_evaluation=False)
    parameter = model.parameter
    metric = model.metric
    
    def uniform(lb, ub, *args, **kwargs):
        return parameter(*args, distribution=shape.Uniform(lb, ub), bounds=(lb, ub), **kwargs)
    
    def default(baseline, *args, **kwargs):
        lb = 0.75*baseline
        ub = 1.25*baseline
        return parameter(*args, distribution=shape.Uniform(lb, ub), bounds=(lb, ub),
                         baseline=baseline, **kwargs)
    
    def default_gwp(baseline, *args, **kwargs):
        lb = 0.90*baseline
        ub = 1.10*baseline
        return parameter(*args, distribution=shape.Uniform(lb, ub), bounds=(lb, ub), 
                         baseline=baseline, **kwargs)
    
    def triangular(lb, mid, ub, *args, **kwargs):
        return parameter(*args, distribution=shape.Triangle(lb, mid, ub), bounds=(lb, ub), **kwargs)
    
    @uniform(40, 70, units='%', kind='coupled')
    def set_bagasse_oil_retention(oil_retention):
        oil_extraction_specification.load_oil_retention(oil_retention / 100.)
    
    def oil_extraction_efficiency_hook(x):
        if number < 0:
            return x
        elif number in (1, 3):
            return 50.0 + x
        elif number in (2, 4):
            return 70.0 + x
    
    @uniform(0., 20, units='%', kind='coupled', 
             hook=oil_extraction_efficiency_hook)
    def set_bagasse_oil_extraction_efficiency(bagasse_oil_extraction_efficiency):
        oil_extraction_specification.load_efficiency(bagasse_oil_extraction_efficiency / 100.)

    # Baseline from Huang's 2016 paper, but distribution more in line with Florida sugarcane harvesting (3-5 months)
    @uniform(4 * 30, 6 * 30, units='day/yr', baseline=180)
    def set_cane_operating_days(cane_operating_days):
        if agile:
            cane_mode.operating_hours = cane_operating_days * 24
        else:
            tea.operating_days = cane_operating_days

    # From Ed Cahoon and Huang 2017
    @uniform(30, 60, units='day/yr', baseline=45)
    def set_sorghum_operating_days(sorghum_operating_days):
        if agile: sorghum_mode.operating_hours = sorghum_operating_days * 24
    
    @default(1761600, units='ton/yr', kind='coupled')
    def set_annual_crushing_capacity(annual_crushing_capacity):
        feedstock.F_mass = kg_per_ton * annual_crushing_capacity / tea.operating_hours

    # USDA ERS historical price data
    @parameter(distribution=ethanol_price_distribution, element=s.ethanol, 
               baseline=1.90, units='USD/gal')
    def set_ethanol_price(price): # Triangular distribution fitted over the past 10 years Sep 2009 to Nov 2020
        s.ethanol.price = price / 2.98668849
        
    # USDA ERS historical price data
    @parameter(distribution=biodiesel_minus_ethanol_price_distribution, element=s.biodiesel, units='USD/gal',
               baseline=2.47, hook=lambda x: s.ethanol.price * 2.98668849 + x)
    def set_biodiesel_price(price): # Triangular distribution fitted over the past 10 years Sep 2009 to March 2021
        s.biodiesel.price = price / 3.3111

    # https://www.eia.gov/energyexplained/natural-gas/prices.php
    @parameter(distribution=natural_gas_price_distribution, element=s.natural_gas, units='USD/cf',
               baseline=4.73)
    def set_natural_gas_price(price): # Triangular distribution fitted over the past 10 years Sep 2009 to March 2021
        BT.natural_gas_price = 51.92624700383502 * price / 1000. 

    # https://www.eia.gov/outlooks/aeo/pdf/00%20AEO2021%20Chart%20Library.pdf
    # Data from historical prices, 2010-2020
    @triangular(0.0583, 0.065, 0.069, units='USD/kWhr',
                baseline=0.0637)
    def set_electricity_price(electricity_price): 
        bst.PowerUtility.price = electricity_price
    
    # 10% is suggested for waste reducing, but 15% is suggested for investment
    @uniform(10., 15., units='%', baseline=10)
    def set_IRR(IRR):
        tea.IRR = IRR / 100.
    
    @uniform(0.10, 0.22, units='USD/kg', element=crude_glycerol)
    def set_crude_glycerol_price(price):
        crude_glycerol.price = price

    @default(0.65, units='USD/kg', element=pure_glycerine)
    def set_pure_glycerol_price(price):
        pure_glycerine.price = price
    
    @default(72, units='hr', element='Saccharification')
    def set_saccharification_reaction_time(reaction_time):
        if abs(number) in (2, 4): saccharification.tau = reaction_time
    
    @default(0.212, units='USD/kg', element='cellulase')
    def set_cellulase_price(price):
        if abs(number) in (2, 4): s.cellulase.price = price
    
    @default(0.02, units='wt. % cellulose', element='cellulase')
    def set_cellulase_loading(cellulase_loading):
        if abs(number) in (2, 4): u.M302.cellulase_loading = cellulase_loading
    
    @default(PRS_cost_item.cost, units='million USD', element='Pretreatment reactor system')
    def set_reactor_base_cost(base_cost):
        PRS_cost_item.cost = base_cost
    
    @uniform(85, 97.5, units='%', element='Pretreatment and saccharification',
             baseline=85)
    def set_cane_glucose_yield(cane_glucose_yield):
        if agile:
            cane_mode.glucose_yield = cane_glucose_yield
        elif abs(number) in (2, 4):
            set_glucose_yield(cane_glucose_yield)
    
    @uniform(79, 97.5, units='%', element='Pretreatment and saccharification',
             baseline=79)
    def set_sorghum_glucose_yield(sorghum_glucose_yield):
        if not agile: return
        sorghum_mode.glucose_yield = sorghum_glucose_yield
        
    @uniform(65, 97.5, units='%', element='Pretreatment and saccharification',
             baseline=65)
    def set_cane_xylose_yield(cane_xylose_yield):
        if agile:
            cane_mode.xylose_yield = cane_xylose_yield
        elif abs(number) in (2, 4):
            set_xylose_yield(cane_xylose_yield)
    
    @uniform(86, 97.5, units='%', element='Pretreatment and saccharification',
             baseline=86)
    def set_sorghum_xylose_yield(sorghum_xylose_yield):
        if not agile: return
        sorghum_mode.xylose_yield = sorghum_xylose_yield
    
    @uniform(90, 95, units='%', element='Cofermenation',
             baseline=90)
    def set_glucose_to_ethanol_yield(glucose_to_ethanol_yield):
        if abs(number) in (2, 4):
            glucose_to_ethanol_yield *= 0.01
            # fermentor.cofermentation[2].X = 0.004 # Baseline
            # fermentor.cofermentation[3].X = 0.006 # Baseline
            # fermentor.loss[0].X = 0.03 # Baseline
            split = np.mean(u.S403.split)
            X1 = split * seed_train.reactions.X[0]
            X2 = split * seed_train.reactions.X[2]
            X3 = (glucose_to_ethanol_yield - X1) / (1 - X1 - X2)
            split = np.mean(u.S403.split)
            X_excess = X3 * 1.0526 - 1
            if X_excess > 0.: breakpoint()
            fermentor.cofermentation.X[0] = X3
            fermentor.cofermentation.X[2] = X3 * 0.0526 # 95% towards ethanol, the other 5% goes towards cell mass
    
    @uniform(50, 95, units='%', element='Cofermenation',
             baseline=50)
    def set_xylose_to_ethanol_yield(xylose_to_ethanol_yield):
        if abs(number) in (2, 4):
            # fermentor.cofermentation[6].X = 0.004 # Baseline
            # fermentor.cofermentation[7].X = 0.046 # Baseline
            # fermentor.cofermentation[8].X = 0.009 # Baseline
            # fermentor.loss[1].X = 0.03 # Baseline
            xylose_to_ethanol_yield *= 0.01
            split = np.mean(u.S403.split)
            X1 = split * seed_train.reactions.X[1]
            X2 = split * seed_train.reactions.X[3]
            X3 = (xylose_to_ethanol_yield - X1) / (1 - X1 - X2)
            X_excess = X3 * 1.0526 - 1
            if X_excess > 0.: breakpoint()
            fermentor.cofermentation.X[1] = X3
            fermentor.cofermentation.X[3] = X3 * 0.0526 # 95% towards ethanol, the other 5% goes towards cell mass

    @uniform(68.5, 137, units='g/L', element='Cofermentation',
             baseline=68.5)
    def set_cofermentation_titer(titer):
        if abs(number) in (2, 4): fermentor.titer = titer

    @uniform(0.951, 1.902, units='g/L', element='Cofermentation',
             baseline=0.951)
    def set_cofermentation_productivity(productivity):
        if abs(number) in (2, 4): fermentor.productivity = productivity

    @default(10, element='oilcane', units='% oil', kind='coupled')
    def set_cane_PL_content(cane_PL_content):
        if agile: cane_mode.PL_content = cane_PL_content / 100.
        else: oil_extraction_specification.PL_content = cane_PL_content / 100.
    
    @default(10, element='oilsorghum', units='% oil', kind='coupled')
    def set_sorghum_PL_content(sorghum_PL_content):
        if agile: sorghum_mode.PL_content = sorghum_PL_content / 100.
    
    @default(10, element='oilcane', units='% oil', kind='coupled')
    def set_cane_FFA_content(cane_FFA_content):
        if agile: cane_mode.FFA_content = cane_FFA_content / 100.
        else: oil_extraction_specification.FFA_content = cane_FFA_content / 100.
    
    @default(10, element='oilsorghum', units='% oil', kind='coupled')
    def set_sorghum_FFA_content(sorghum_FFA_content):
        if agile: sorghum_mode.FFA_content = sorghum_FFA_content / 100. 

    @uniform(5., 15., element='oilcane', units='dry wt. %', kind='coupled')
    def set_cane_oil_content(cane_oil_content):
        if agile:
            cane_mode.oil_content = cane_oil_content / 100.
        else:
            oil_extraction_specification.load_oil_content(cane_oil_content / 100.)

    @uniform(-3., 0., element='oilsorghum', units='dry wt. %', kind='coupled',
             baseline=0.)
    def set_relative_sorghum_oil_content(relative_sorghum_oil_content):
        if agile:
            sorghum_mode.oil_content = cane_mode.oil_content + relative_sorghum_oil_content / 100.

    @default(23, units='% oil', kind='coupled', name='TAG to FFA conversion')
    def set_TAG_to_FFA_conversion(TAG_to_FFA_conversion):
        if number in (1, 3):
            u.R301.oil_reaction.X[0] = TAG_to_FFA_conversion / 100.
        elif number in (2, 4):
            u.R401.oil_reaction.X[0] = TAG_to_FFA_conversion / 100.
    
    @default_gwp(feedstock.characterization_factors[GWP], name='GWP', 
             element=feedstock, units='kg*CO2-eq/kg')
    def set_feedstock_GWP(value):
        if number > 0:
            feedstock.characterization_factors[GWP] = value
    
    @default_gwp(s.methanol.characterization_factors[GWP], name='GWP', 
                 element=s.methanol, units='kg*CO2-eq/kg')
    def set_methanol_GWP(value):
        s.methanol.characterization_factors[GWP] = value
    
    # @default(crude_glycerol.characterization_factors[GWP], name='GWP', 
    #          element=crude_glycerol, units='kg*CO2-eq/kg')
    # def set_crude_glycerol_GWP(value):
    #     crude_glycerol.characterization_factors[GWP] = value
    
    @default_gwp(s.pure_glycerine.characterization_factors[GWP], name='GWP', 
                 element=pure_glycerine, units='kg*CO2-eq/kg')
    def set_pure_glycerine_GWP(value):
        s.pure_glycerine.characterization_factors[GWP] = value
    
    @default_gwp(s.cellulase.characterization_factors[GWP], name='GWP', 
                 element=s.cellulase, units='kg*CO2-eq/kg')
    def set_cellulase_GWP(value):
        s.cellulase.characterization_factors[GWP] = value * 0.02
    
    @default_gwp(s.natural_gas.characterization_factors[GWP], name='GWP', 
                 element=s.natural_gas, units='kg*CO2-eq/kg')
    def set_natural_gas_GWP(value):
        for ng in natural_gas_streams:
            ng.characterization_factors[GWP] = value
    
    s.natural_gas.phase = 'g'
    s.natural_gas.set_property('T', 60, 'degF')
    s.natural_gas.set_property('P', 14.73, 'psi')
    original_value = s.natural_gas.imol['CH4']
    s.natural_gas.imass['CH4'] = 1 
    V_ng = s.natural_gas.get_total_flow('ft3/hr')
    s.natural_gas.imol['CH4'] = original_value
    
    if agile:
        feedstock_flow = lambda: sys.flow_rates[feedstock] / kg_per_ton # ton/yr
        biodiesel_flow = lambda: sys.flow_rates.get(s.biodiesel, 0.) / 3.3111 # gal/yr
        ethanol_flow = lambda: sys.flow_rates[s.ethanol] / 2.98668849 # gal/yr
        natural_gas_flow = lambda: sum([sys.flow_rates[i] for i in natural_gas_streams]) * V_ng # cf/yr
        crude_glycerol_flow = lambda: sys.flow_rates.get(s.crude_glycerol, 0.) # kg/yr
        
        @sys.operation_metric(annualize=True)
        def direct_nonbiogenic_emissions(mode):
            return sum([i.F_mol for i in natural_gas_streams]) * chemicals.CO2.MW
        
    else:
        feedstock_flow = lambda: sys.operating_hours * feedstock.F_mass / kg_per_ton # ton/yr
        biodiesel_flow = lambda: sys.operating_hours * s.biodiesel.F_mass / 3.3111 # gal/yr
        ethanol_flow = lambda: sys.operating_hours * s.ethanol.F_mass / 2.98668849 # gal/yr
        crude_glycerol_flow = lambda: sys.operating_hours * s.crude_glycerol.F_mass # kg/yr
        natural_gas_flow = lambda: sum([i.F_mass for i in natural_gas_streams]) * sys.operating_hours * V_ng # cf/yr
        direct_nonbiogenic_emissions = lambda: sum([i.F_mol for i in natural_gas_streams]) * chemicals.CO2.MW * sys.operating_hours
    electricity = lambda: sys.operating_hours * sys.power_utility.rate
    
    sys.define_process_impact(
        key=GWP,
        name='Direct non-biogenic emissions',
        basis='kg',
        inventory=direct_nonbiogenic_emissions,
        CF=1.,
    )
    
    dct['flows'] = {
        'feedstock': feedstock_flow,
        'biodiesel': biodiesel_flow,
        'ethanol': ethanol_flow,
        'natural_gas': natural_gas_flow,
        'crude_glycerol': crude_glycerol_flow,
        'electricity': electricity
    }
    @metric(units='USD/ton')
    def MFPP():
        price = tea.solve_price(feedstock)
        return kg_per_ton * price
    
    @metric(units='ton/yr')
    def feedstock_consumption():
        return feedstock_flow()
    
    @metric(units='Gal/ton')
    def biodiesel_production():
        return biodiesel_flow() / feedstock_consumption.get()
    
    @metric(units='Gal/ton')
    def ethanol_production():
        return ethanol_flow() / feedstock_consumption.get()
    
    @metric(units='kWhr/ton')
    def electricity_production():
        value = - electricity() / feedstock_consumption.get()
        if value < 0.: value = 0.
        return value
    
    @metric(units='cf/ton')
    def natural_gas_consumption():
        value = natural_gas_flow() / feedstock_consumption.get()
        return value
    
    @metric(units='10^6*USD')
    def TCI():
        return tea.TCI / 1e6 # 10^6*$
    
    @metric(units='%')
    def heat_exchanger_network_error():
        return HXN.energy_balance_percent_error if HXN else 0.    

    def GWP_displacement(): # Cradle to gate
        GWP_total = sys.get_total_feeds_impact(GWP) + min(electricity(), 0) * GWP_characterization_factors['Electricity'] # kg CO2 eq. / yr
        sales = (
            biodiesel_flow() * mean_biodiesel_price
            + ethanol_flow() * mean_ethanol_price
            + crude_glycerol_flow() * mean_glycerol_price
        )
        GWP_per_USD = GWP_total / sales
        return {
            'Ethanol': GWP_per_USD * mean_ethanol_price,
            'Biodiesel': GWP_per_USD * mean_biodiesel_price,
            'Crude glycerol': GWP_per_USD * mean_glycerol_price,
        }
    dct['GWP_displacement'] = GWP_displacement

    @metric(name='GWP', element='Economic allocation', units='kg*CO2*eq / USD')
    def GWP_economic(): # Cradle to gate
        GWP_material = sys.get_total_feeds_impact(GWP) # kg CO2 eq. / yr
        GWP_emissions = sys.get_process_impact(GWP) # kg CO2 eq. / yr
        sales = (
            biodiesel_flow() * mean_biodiesel_price
            + ethanol_flow() * mean_ethanol_price
            + crude_glycerol_flow() * mean_glycerol_price
            + max(-electricity(), 0) * mean_electricity_price
        )
        return (GWP_material + GWP_emissions) / sales

    @metric(name='Ethanol GWP', element='Economic allocation', units='kg*CO2*eq / gal')
    def GWP_ethanol(): # Cradle to gate
        return GWP_economic.get() * mean_ethanol_price
    
    @metric(name='Biodiesel GWP', element='Economic allocation', units='kg*CO2*eq / gal')
    def GWP_biodiesel(): # Cradle to gate
        if number > 0:
            return GWP_economic.get() * mean_biodiesel_price
        else:
            return 0.
    
    @metric(name='Crude glycerol GWP', element='Economic allocation', units='kg*CO2*eq / kg')
    def GWP_crude_glycerol(): # Cradle to gate
        if number > 0:
            return GWP_economic.get() * mean_glycerol_price
        else:
            return 0.
    
    @metric(name='Electricity GWP', element='Economic allocation', units='kg*CO2*eq / MWhr')
    def GWP_electricity(): # Cradle to gate
        if abs(number) in (1, 3):
            return GWP_economic.get() * mean_electricity_price * 1000.
        else:
            return 0.

    @metric(name='Ethanol GWP', element='Displacement allocation', units='kg*CO2*eq / gal')
    def GWP_ethanol_displacement(): # Cradle to gate
        GWP_material = sys.get_total_feeds_impact(GWP)
        GWP_electricity_production = GWP_characterization_factors['Electricity'] * electricity_production.get() * feedstock_consumption.get()
        GWP_coproducts = sys.get_total_products_impact(GWP)
        GWP_emissions = sys.get_process_impact(GWP) # kg CO2 eq. / yr
        GWP_total = GWP_material + GWP_emissions - GWP_electricity_production - GWP_coproducts # kg CO2 eq. / yr
        return GWP_total / (ethanol_production.get() * feedstock_consumption.get())
    
    # import thermosteam as tmo
    # glycerol = tmo.Chemical('Glycerol')
    # ethanol = tmo.Chemical('Ethanol')
    # glycerol_GGE = 0.80 * (glycerol.LHV / glycerol.MW) / 121300 # 0.1059 GGE / kg crude-glycerol
    
    @metric(name='Biofuel GWP', element='Energy allocation', units='kg*CO2*eq / GGE')
    def GWP_biofuel_allocation(): # Cradle to gate
        GWP_material = sys.get_total_feeds_impact(GWP)
        GWP_emissions = sys.get_process_impact(GWP) # kg CO2 eq. / yr
        GWP_total = GWP_material + GWP_emissions # kg CO2 eq. / yr
        GGE_biodiesel_annual = (biodiesel_production.get() * feedstock_consumption.get()) / 0.9536
        GGE_ethanol_annual = (ethanol_production.get() * feedstock_consumption.get()) / 1.5
        GEE_electricity_production = max(-electricity() * 3600 / 114000, 0.) 
        GEE_crude_glycerol = crude_glycerol_flow() * 0.1059
        return GWP_total / (GGE_biodiesel_annual + GGE_ethanol_annual + GEE_electricity_production + GEE_crude_glycerol)
    
    @metric(name='Ethanol GWP', element='Energy allocation', units='kg*CO2*eq / gal')
    def GWP_ethanol_allocation(): # Cradle to gate
        return GWP_biofuel_allocation.get() / 1.5
    
    @metric(name='Biodiesel GWP', element='Energy allocation', units='kg*CO2*eq / gal')
    def GWP_biodiesel_allocation(): # Cradle to gate
        if number > 0:
            return GWP_biofuel_allocation.get() / 0.9536
        else:
            return 0.
    
    @metric(name='Crude-glycerol GWP', element='Energy allocation', units='kg*CO2*eq / kg')
    def GWP_crude_glycerol_allocation(): # Cradle to gate
        if number > 0:
            return GWP_biofuel_allocation.get() * 0.1059
        else:
            return 0.

    @metric(units='USD/ton')
    def MFPP_derivative():
        if number < 0: return 0.
        if _derivative_disabled: return np.nan
        if agile:
            cane_mode.oil_content += 0.01
            sorghum_mode.oil_content += 0.01
        else:
            oil_extraction_specification.load_oil_content(oil_extraction_specification.oil_content + 0.01)
        sys.simulate()  
        # value = (kg_per_ton * tea.solve_price(feedstock) - MFPP.cache)
        # feedstock.price = tea.solve_price(feedstock)
        # print('AFTER')
        # print('MFPP', kg_per_ton * tea.solve_price(feedstock))
        # print('VOC', tea.VOC / 1e3)
        # print('TCI', tea.TCI / 1e6)
        # print('sales', tea.sales / 1e3)
        # print('NPV', tea.NPV)
        return MFPP.difference()
    
    @metric(units='Gal/ton')
    def biodiesel_production_derivative():
        if number < 0: return 0.
        if _derivative_disabled: return np.nan
        return biodiesel_production.difference()
    
    @metric(units='Gal/ton')
    def ethanol_production_derivative():
        if number < 0: return 0.
        if _derivative_disabled: return np.nan
        return ethanol_production.difference()
    
    @metric(units='kWhr/ton')
    def electricity_production_derivative():
        if number < 0: return 0.
        if _derivative_disabled: return np.nan
        return electricity_production.difference()
    
    @metric(units='cf/ton')
    def natural_gas_consumption_derivative():
        if number < 0: return 0.
        if _derivative_disabled: return np.nan
        # print('natural gas production derivative', value)
        return natural_gas_consumption.difference()
    
    @metric(units='10^6*USD')
    def TCI_derivative():
        if number < 0: return 0.
        if _derivative_disabled: return np.nan
        return TCI.difference()
    
    @metric(name='GWP derivative', element='Economic allocation', units='kg*CO2*eq / USD')
    def GWP_economic_derivative(): # Cradle to gate
        if number < 0: return 0.
        if _derivative_disabled: return np.nan
        return GWP_economic.difference()

    @metric(name='Ethanol GWP derivative', element='Ethanol', units='kg*CO2*eq / gal')
    def GWP_ethanol_derivative(): # Cradle to gate
        return GWP_economic_derivative.get() * mean_ethanol_price
    
    @metric(name='Biodiesel GWP derivative', element='Biodiesel', units='kg*CO2*eq / gal')
    def GWP_biodiesel_derivative(): # Cradle to gate
        if number > 0:
            return GWP_economic_derivative.get() * mean_biodiesel_price
        else:
            return 0.
    
    @metric(name='Crude glycerol GWP derivative', element='Crude glycerol', units='kg*CO2*eq / kg')
    def GWP_crude_glycerol_derivative(): # Cradle to gate
        if number > 0:
            return GWP_economic_derivative.get() * mean_glycerol_price
        else:
            return 0.
    
    @metric(name='Electricity GWP derivative', element='Electricity', units='kg*CO2*eq / MWhr')
    def GWP_electricity_derivative(): # Cradle to gate
        if abs(number) in (1, 3):
            return GWP_economic_derivative.get() * mean_electricity_price * 1000.
        else:
            return 0.
    
    # @metric(units='MMGGE/yr')
    # def productivity():
    #     GGE = (ethanol_flow() / 1.5
    #         + biodiesel_flow() / 0.9536
    #         - electricity() * 3600 / 131760
    #         - natural_gas_flow() / 126.67)
    #     return GGE / 1e6
    
    # Single point evaluation for detailed design results
    def set_baseline(p, x):
        p.setter(x)
        p.baseline = x
    
    if abs(number) in (2, 4):
        if enhanced_cellulosic_performance:
            set_baseline(set_sorghum_glucose_yield, 97.5)
            set_baseline(set_sorghum_xylose_yield, 97.5)
            set_baseline(set_cane_glucose_yield, 97.5)
            set_baseline(set_cane_xylose_yield, 97.5)
            set_baseline(set_glucose_to_ethanol_yield, 95)
            set_baseline(set_xylose_to_ethanol_yield, 95)
            set_baseline(set_cofermentation_titer, 120.)
            set_baseline(set_cofermentation_productivity, 2.0)
        else:
            set_baseline(set_sorghum_glucose_yield, 79)
            set_baseline(set_sorghum_xylose_yield, 86)
            set_baseline(set_cane_glucose_yield, 91.0)
            set_baseline(set_cane_xylose_yield, 97.5)
            set_baseline(set_glucose_to_ethanol_yield, 90)
            set_baseline(set_xylose_to_ethanol_yield, 42)
    if number in (1, 3) and enhanced_biodiesel_production:
        set_baseline(set_cane_oil_content, 15)
        set_baseline(set_bagasse_oil_extraction_efficiency, oil_extraction_efficiency_hook(20))
        set_baseline(set_bagasse_oil_retention, 40)
    else:
        set_baseline(set_cane_oil_content, 5)
        set_baseline(set_bagasse_oil_extraction_efficiency, oil_extraction_efficiency_hook(0.))
        set_baseline(set_bagasse_oil_retention, 70)
    set_baseline(set_ethanol_price, mean_ethanol_price) 
    set_baseline(set_crude_glycerol_price, mean_glycerol_price)
    set_baseline(set_biodiesel_price, mean_biodiesel_price)
    set_baseline(set_natural_gas_price, mean_natural_gas_price)
    set_baseline(set_electricity_price, mean_electricity_price)
    if number > 0:
        set_baseline(set_cane_PL_content, 10)
        set_baseline(set_cane_FFA_content, 10)
    # set_fermentation_solids_loading(20) # Same as Humbird
    # set_feedstock_oil_content(10) # Consistent with SI of Huang's 2016 paper
    # set_ethanol_price(2.356) # Consistent with Huang's 2016 paper
    # set_biodiesel_price(4.569) # Consistent with Huang's 2016 paper
    # set_natural_gas_price(4.198) # Consistent with Humbird's 2012 paper
    # set_electricity_price(0.0572) # Consistent with Humbird's 2012 paper
    # set_operating_days(200) # Consistent with Huang's 2016 paper
    
    for i in sys.units:
        if isinstance(i, bst.MultiEffectEvaporator): i.flash = False
    
    for i in model._parameters:
        dct[i.setter.__name__] = i
    for i in model._metrics:
        dct[i.getter.__name__] = i
    if cache is not None: cache[key] = dct.copy()
    
    ## Simulation
    HXN.force_ideal_thermo = True
    HXN.cache_network = True
    HXN.avoid_recycle = True
    try: 
        sys.simulate()
    except Exception as e:
        raise e
    if reduce_chemicals:
        oilcane_sys.reduce_chemicals()
    oilcane_sys._load_stream_links()
    HXN.simulate()

# DO NOT DELETE: For removing ylabel and yticklabels and combine plots
# import biorefineries.oilcane as oc
# import matplotlib.pyplot as plt
# oc.plot_configuration_breakdown('O2')
# ax, *_ = plt.gcf().get_axes()
# yticks = ax.get_yticks()
# plt.yticks(yticks, ['']*len(yticks))
# plt.ylabel('')
# plt.show()

# DO NOT DELETE: For better TEA tickmarks
# import biorefineries.oilcane as oc
# import numpy as np
# oc.plot_monte_carlo(derivative=True, comparison=False,
#     tickmarks=np.array([
#         [-3.0, -1.5, 0, 1.5, 3.0, 4.5],
#         [-6, -3, 0, 3, 6, 9],
#         [-2.25, -1.5, -0.75, 0, 0.75, 1.5],
#         [-10, 0, 10, 20, 30, 40],
#         [-300, -225, -150, -75, 0, 75]
#     ]),
#     labels=['Conventional', 'Cellulosic']
# )

# DO NOT DELETE: For better LCA tickmarks
# import biorefineries.oilcane as oc
# import numpy as np
# oc.plot_monte_carlo(kind='LCA', comparison=True, agile=False,
#     tickmarks=np.array([
#         [-2, -1, 0, 1, 2, 3, 4, 5],
#         [-2, 0, 2, 4, 6, 8, 10],
#         [-1, 0, 1, 2, 3, 4, 5],
#         [-150, -75, -50, -25, 0., 75, 150, 225, 300],
#     ]),
# )

# DO NOT DELETE: For SI Monte Carlo
# import biorefineries.oilcane as oc
# import numpy as np
# oc.plot_monte_carlo(
#     absolute=True, comparison=False,
#     labels=['Sugarcane\nconventional', 'Oilcane\nconventional',
#             'Sugarcane\ncellulosic', 'Oilcane\ncellulosic',
#             'Sugarcane\nconventional\nagile', 'Oilcane\nconventional\nagile',
#             'Sugarcane\ncellulosic\nagile', 'Oilcane\ncellulosic\nagile'],
# )

# # DO NOT DELETE: For GWP tables
# from biorefineries import oilcane as oc
# import biosteam as bst
# def get_sys(ID):
#     oc.load(ID)
#     return oc.sys
# def get_ethanol(ID):
#     oc.load(ID)
#     return oc.ethanol
# IDs = ('S1', 'S2', 'O1', 'O2')
# systems = [get_sys(i) for i in IDs]
# items = [get_ethanol(i) for i in IDs]
# bst.settings.define_impact_indicator('GWP', 'kg*CO2e')
# bst.report.lca_table_displacement_allocation(systems, 'GWP', items, 'ethanol', system_names=IDs)

# # Calculate xylose conversion based on net conversion of sugars
# import biosteam as bst
# import biorefineries.oilcane as oc
# oc.load('L2')
# feed = bst.Stream.sum(oc.R401.ins)
# total_glucose, total_xylose = feed.imass['Glucose', 'Xylose'] + feed.imass['Glucan', 'Xylan'] * 1.11
# total_sugars = total_glucose + total_xylose
# glucose = 0.91 * total_glucose
# xylose = total_xylose
# sugar_conversion = 0.83 * 0.95 
# xylose_conversion = (sugar_conversion * total_sugars - 0.95 * glucose) / xylose
# print(xylose_conversion)



# -*- coding: utf-8 -*-
"""
Created on Tue Oct  5 03:46:23 2021

@author: yrc2
"""
import numpy as np
# Switchgrass composition

# Fiber
Arabinan = [3.7, 4.0, 4.0, 3.8]
Galactan = [1.3, 1.5, 1.5, 1.5]
Glucan = [37, 37.1, 38, 38.9]
Xylan = [28.8, 29.0, 30.1, 30.0]
Mannan = [0.9, 0.9, 0.7, 0.8]
Lignin = [22.7, 23.5, 24.0, 24.9]
Ash = [2.1, 2.5, 2.2, 2.2]
Extractives = [11, 11, 12, 13]

composition = np.array(
    [Arabinan, Galactan, Glucan,
     Xylan, Mannan, Lignin, Ash,
     Extractives]
)

composition = composition.sum(axis=1)
composition /= composition.sum()
IDs = ['Arabinan', 'Galactan', 'Glucan',
       'Xylan', 'Mannan', 'Lignin', 'Ash',
       'Extractives']
dw = 0.8
switchgrass_composition_dct = {i: float(j) * dw for i,j in zip(IDs, composition)}
acetate = 0.008966261808367074
switchgrass_composition_dct['Extractives'] -= acetate
switchgrass_composition_dct['Acetate'] = acetate
switchgrass_composition_dct['Water'] = 0.2


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# Bioindustrial-Park: BioSTEAM's Premier Biorefinery Models and Results
# Copyright (C) 2020-2021, Yalin Li <yalinli2@illinois.edu>,
# Sarang Bhagwat <sarangb2@illinois.edu>, and Yoel Cortes-Pena (this biorefinery)
#
# This module is under the UIUC open-source license. See
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.


# %%

# =============================================================================
# Setup
# =============================================================================

import numpy as np
import pandas as pd
import biosteam as bst
from biosteam.utils import TicToc
from biosteam.plots import plot_montecarlo_across_coordinate
from biorefineries.lactic import create_model, \
    SSCF_flowsheet, SSCF_funcs, SHF_flowsheet, SHF_funcs


# %%

def evaluate_uncertainties(kind='SSCF', seed=None, N_simulation=1000,
                           sampling_rule='L',
                           percentiles = [0, 0.05, 0.25, 0.5, 0.75, 0.95, 1],
                           if_plot=True, report_name='1_full_evaluation.xlsx'):
    if 'SSCF' in str(kind).upper():
        flowsheet = SSCF_flowsheet
        funcs = SSCF_funcs
    elif 'SHF' in str(kind).upper():
        flowsheet = SHF_flowsheet
        funcs = SHF_funcs
    else:
        raise ValueError(f'kind can only be "SSCF" or "SHF", not {kind}.')

    simulate_get_MPSP = funcs['simulate_get_MPSP']
    simulate_get_MPSP()
    model_dct = create_model(kind)
    model = model_dct['model']

    # =============================================================================
    # Evaluate and organize results for Monte Carlo analysis
    # =============================================================================

    # Initiate a timer
    timer = TicToc('timer')
    timer.tic()

    # Set seed to make sure each time the same set of random numbers will be used
    if seed:
        np.random.seed(seed)
    samples = model.sample(N=N_simulation, rule=sampling_rule)
    model.load_samples(samples)

    baseline_initial = model.metrics_at_baseline()
    baseline = pd.DataFrame(data=np.array([[i for i in baseline_initial.values],]),
                            columns=baseline_initial.keys())

    model.evaluate()

    # Baseline results
    baseline_end = model.metrics_at_baseline()
    baseline = baseline.append(baseline_end, ignore_index=True)
    baseline.index = ('initial', 'end')
    # baseline.to_excel('0_baseline.xlsx')

    # Parameters
    parameters = model.get_parameters()
    index_parameters = len(model.get_baseline_sample())
    parameter_values = model.table.iloc[:, :index_parameters].copy()

    # TEA results
    index_TEA = index_parameters + model_dct['index_TEA']
    TEA_results = \
        model.table.iloc[:, index_parameters:index_TEA].copy()
    TEA_percentiles = TEA_results.quantile(q=percentiles)

    # MPSP across IRRs
    index_IRR = index_parameters + model_dct['index_IRR']
    IRR_results = \
        model.table.iloc[:, index_TEA:index_IRR].copy()
    IRR_percentiles = IRR_results.quantile(q=percentiles)

    # LCA_results
    LCA_results = \
        model.table.iloc[:, index_IRR::].copy()
    LCA_percentiles = LCA_results.quantile(q=percentiles)

    # Spearman's rank correlation
    spearman_metrics = model.metrics[0:2] + model.metrics[6:8] + \
        model.metrics[index_IRR:index_IRR+2]

    model.table = model.table.dropna()
    spearman_parameters = parameters
    spearman_results = model.spearman_r(spearman_parameters, spearman_metrics)[0]
    spearman_results.columns = pd.Index([i.name_with_units for i in spearman_metrics])

    # Calculate the cumulative probabilitie of each parameter
    probabilities = {}
    for i in range(index_parameters):
        p = parameters[i]
        p_values = parameter_values.iloc[:, 2*i]
        probabilities[p.name] = p.distribution.cdf(p_values)
        parameter_values.insert(loc=2*i+1,
                          column=(parameter_values.iloc[:, 2*i].name[0], 'Probability'),
                          value=probabilities[p.name],
                          allow_duplicates=True)

    run_number = samples.shape[0]

    # =============================================================================
    # Evaluate the min/max of one parameter each time to ensure the parameter can
    # independently affect the system
    # =============================================================================

    p_values = [[], [], []]
    MPSPs = [[], [], []]
    GWPs = [[], [], []]
    FECs = [[], [], []]

    bst.speed_up()
    for p in parameters:
        p_min = p.distribution.lower.item()
        p_max = p.distribution.upper.item()
        p_baseline = p.baseline
        p_value = (p_min, p_max, p_baseline)
        p.system = flowsheet.system.lactic_sys
        for i in range(len(p_value)):
            p.setter(p_value[i])
            p_values[i].append(p_value[i])
            MPSP = simulate_get_MPSP()
            MPSPs[i].append(MPSP)
            GWPs[i].append(funcs['get_GWP']())
            FECs[i].append(funcs['get_FEC']())
            run_number += 1

    MPSP_baseline = np.asarray(MPSPs[2])
    MPSP_min_diff = np.asarray(MPSPs[0]) - MPSP_baseline
    MPSP_max_diff = np.asarray(MPSPs[1]) - MPSP_baseline

    GWP_baseline = np.asarray(GWPs[2])
    GWP_min_diff = np.asarray(GWPs[0]) - GWP_baseline
    GWP_max_diff = np.asarray(GWPs[1]) - GWP_baseline

    FEC_baseline = np.asarray(FECs[2])
    FEC_min_diff = np.asarray(FECs[0]) - FEC_baseline
    FEC_max_diff = np.asarray(FECs[1]) - FEC_baseline

    one_p_df = pd.DataFrame({
        ('Parameter', 'Name'): [i.name_with_units for i in parameters],
        ('Parameter', 'Baseline'): p_values[2],
        ('Parameter', 'Min'): p_values[0],
        ('Parameter', 'Max'): p_values[1],
        ('MPSP [$/kg]', 'MPSP baseline'): MPSP_baseline,
        ('MPSP [$/kg]', 'MPSP min'): MPSPs[0],
        ('MPSP [$/kg]', 'MPSP min diff'): MPSP_min_diff,
        ('MPSP [$/kg]', 'MPSP max'): MPSPs[1],
        ('MPSP [$/kg]', 'MPSP max diff'): MPSP_max_diff,
        ('GWP [kg CO2-eq/kg]', 'GWP baseline'): GWP_baseline,
        ('GWP [kg CO2-eq/kg]', 'GWP min'): GWPs[0],
        ('GWP [kg CO2-eq/kg]', 'GWP min diff'): GWP_min_diff,
        ('GWP [kg CO2-eq/kg]', 'GWP max'): GWPs[1],
        ('GWP [kg CO2-eq/kg]', 'GWP max diff'): GWP_max_diff,
        ('FEC [MJ/kg]', 'FEC baseline'): FEC_baseline,
        ('FEC [MJ/kg]', 'FEC min'): FECs[0],
        ('FEC [MJ/kg]', 'FEC min diff'): FEC_min_diff,
        ('FEC [MJ/kg]', 'FEC max'): FECs[1],
        ('FEC [MJ/kg]', 'FEC max diff'): FEC_max_diff,
        })

    time = timer.elapsed_time / 60
    print(f'\nSimulation time for {run_number} runs is: {time:.1f} min')


    if if_plot:
        IRR_plot_indices = [metric.index for metric in model.metrics
                            if 'IRR' in metric.index[0] and 'MPSP' in metric.index[1]]
        IRR_plot_data = IRR_results[IRR_plot_indices].copy()
        IRR_plot_data.columns = model_dct['IRRs'].copy()
        IRR_plot_y = IRR_plot_data.sort_index(axis=1)
        IRR_plot_y = IRR_plot_y.dropna()
        IRR_plot_x = model_dct['IRRs'].copy()
        IRR_plot_x.sort()
        plot_montecarlo_across_coordinate(IRR_plot_x, IRR_plot_y)

    if report_name:
        with pd.ExcelWriter(report_name) as writer:
            parameter_values.to_excel(writer, sheet_name='Parameters')
            TEA_results.to_excel(writer, sheet_name='TEA results')
            TEA_percentiles.to_excel(writer, sheet_name='TEA percentiles')
            IRR_results.to_excel(writer, sheet_name='IRR results')
            IRR_percentiles.to_excel(writer, sheet_name='IRR percentiles')
            LCA_results.to_excel(writer, sheet_name='LCA results')
            LCA_percentiles.to_excel(writer, sheet_name='LCA percentiles')
            spearman_results.to_excel(writer, sheet_name='Spearman')
            one_p_df.to_excel(writer, sheet_name='One-parameter')
            model.table.to_excel(writer, sheet_name='Raw data')

evaluate_uncertainties(kind='SSCF', seed=None, N_simulation=10, sampling_rule='L',
                        percentiles = [0, 0.05, 0.25, 0.5, 0.75, 0.95, 1],
                        if_plot=True, report_name='1_full_evaluation.xlsx')


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.

import flexsolve as flx
from scipy.optimize import minimize_scalar
import numpy as np
from thermosteam.exceptions import InfeasibleRegion
from biorefineries.lipidcane import set_lipid_fraction as set_oil_fraction
import flexsolve as flx
import thermosteam as tmo
import biosteam as bst

__all__ = ('OilExtractionSpecification', 'MockExtractionSpecification')

def evaluate_across_oil_content(spec, efficiency, oil_retention, 
                                  metrics, oil_content):
    spec.load_specifications(efficiency=efficiency, oil_retention=oil_retention)
    return spec._evaluate_across_oil_content(metrics, oil_content)
    
evaluate_across_oil_content = np.vectorize(
    evaluate_across_oil_content, 
    excluded=['spec', 'metrics', 'oil_content'],
    signature='(),(),(),(m),(p)->(m,p)'
)

def evaluate_across_oil_retention(spec, efficiency, oil_content, 
                                    metrics, oil_retention):
    spec.load_specifications(efficiency=efficiency, oil_content=oil_content)
    return spec._evaluate_across_oil_retention(metrics, oil_retention)
    
evaluate_across_oil_retention = np.vectorize(
    evaluate_across_oil_retention, 
    excluded=['spec', 'metrics', 'oil_retention'],
    signature='(),(),(),(m),(p)->(m,p)'
)       

class MockExtractionSpecification:
    
    def load_oil_content(self, oil_content):
        pass
      
    def load_efficiency(self, efficiency):
        pass
    
    def load_oil_retention(self, oil_retention):
        pass
    
    def load_specifications(self, 
            efficiency=None,
            oil_retention=None,
            oil_content=None,
        ):
        pass

class OilExtractionSpecification:
    """
    Create a OilExtractionSpecification object for setting process 
    specifications related to oil extraction.
    
    Parameters
    ----------
    system : System
        System associated to feedstocks.
    feedstocks : Stream
        Oilcane feedstocks.
    isplit_efficiency : ChemicalIndexer
        Defines extraction efficiency as a material split.
    isplit_oil_retention : ChemicalIndexer
        Defines bagasse oil retention as a material split.
    oil_content : float 
        Oil content of feedstocks [dry wt. %].
        
    Notes
    -----
    Use the `load_specifications` method to load specifications.
    Setting attributes (i.e. yield_, titer, productivity, production) sets 
    defaults for this method, but does not actually load any specifications.
    
    """
    
    __slots__ = (
        'system',
        'feedstocks',
        'isplit_efficiency',
        'isplit_oil_retention',
        'efficiency',
        'oil_retention',
        'oil_content',
        'locked_oil_content',
        'isplit_efficiency_is_reversed',
        'FFA_content',
        'PL_content',
    )
    
    def __init__(self, system, feedstocks, isplit_efficiency, isplit_oil_retention, 
                 isplit_efficiency_is_reversed=False, efficiency=0.9, 
                 oil_retention=0.9, oil_content=0.1, FFA_content=0.1,
                 PL_content=0.1):
        self.system = system #: [System] System associated to feedstocks
        self.feedstocks = feedstocks #: Stream Oil feedstocks
        self.isplit_efficiency = isplit_efficiency #: [ChemicalIndexer] Defines extraction efficiency as a material split.
        self.isplit_oil_retention = isplit_oil_retention #: [ChemicalIndexer] Defines bagasse oil retention as a material split.
        self.isplit_efficiency_is_reversed = isplit_efficiency_is_reversed #: [bool] Whether oil extraction efficiency is 1 - split.
        self.efficiency = efficiency #: [float] Oil extraction efficiency b
        self.oil_retention = oil_retention #: [float] Oil extraction oil retention
        self.oil_content = oil_content #: [float] Oil content of feedstocks [dry wt. %].
        self.locked_oil_content = False
        self.FFA_content = FFA_content
        self.PL_content = PL_content
        
    def MFPP(self):
        return self.system.TEA.solve_price(self.feedstocks)
        
    def dMFPP_over_doil_content_at_efficiency(self, efficiency, doil_content=0.01):
        oil_content = self.oil_content
        system = self.system
        self.load_efficiency(efficiency)   
        self.load_oil_content(oil_content + doil_content)
        system.simulate()
        MFPP1 = self.MFPP()
        self.load_oil_content(oil_content)
        system.simulate()
        MFPP0 = self.MFPP()
        return (MFPP1 - MFPP0) / doil_content
        
    def solve_MFPP_inflection(self, oil_retention=None, oil_content=None):
        if oil_content is not None: self.load_oil_content(oil_content)
        if oil_retention is not None: self.load_oil_retention(oil_retention)
        f = self.dMFPP_over_doil_content_at_efficiency
        x0 = 0.1
        x1 = 1.
        y0 = f(x0)
        y1 = f(x1)
        if not y0 < 0. < y1: return np.nan
        return flx.IQ_interpolation(f, 0.1, 1., y0, y1, xtol=1e-4, ytol=1e-4, checkiter=False,
                                    maxiter=10)
    
    def load_oil_content(self, oil_content):
        if self.locked_oil_content: return
        for i in self.feedstocks: 
            set_oil_fraction(
                oil_content, i, PL_fraction=self.PL_content, 
                FFA_fraction=self.FFA_content
            )
        self.oil_content = oil_content
      
    def load_efficiency(self, efficiency):
        if self.isplit_efficiency is None: return
        self.efficiency = efficiency
        if self.isplit_efficiency_is_reversed:
            self.isplit_efficiency['Oil'] = 1. - efficiency
        else:
            self.isplit_efficiency['Oil'] = efficiency
    
    def load_oil_retention(self, oil_retention):
        if self.isplit_efficiency is None: return
        self.oil_retention = oil_retention
        self.isplit_oil_retention['Oil'] = oil_retention    
    
    def load_specifications(self, 
            efficiency=None,
            oil_retention=None,
            oil_content=None,
        ):
        """
        Load oil extraction specifications.

        Parameters
        ----------
        efficiency : float, optional
            Oil extraction efficiency a.
        oil_retention : float, optional
            Oil extraction efficiency b.
        oil_content : float, optional
            Oil content of feedstocks [dry wt. %].

        """
        if efficiency is None: efficiency = self.efficiency
        if oil_retention is None: oil_retention = self.oil_retention
        if oil_content is None: oil_content = self.oil_content
        self.load_oil_retention(oil_retention)
        self.load_efficiency(efficiency)
        self.load_oil_content(oil_content)

    def _evaluate_across_oil_content(self, metrics, oil_content):
        """
        Evaluate metrics across oil content and return an array with the all
        metric results.
        
        Parameters
        ----------
        metrics : Iterable[Callable; M elements]
            Should return a number given no parameters.
        oil_content : array_like[P elements]
            Oil contents to evaluate.
        
        Returns
        -------
        results : array[M x P]
            All metric results.
        
        """
        M = len(metrics)
        P = len(oil_content)
        data = np.zeros([M, P])
        for i in range(P):
            self.load_oil_content(oil_content[i])
            self.system.simulate()
            data[:, i] = [j() for j in metrics]
        return data
    
    def _evaluate_across_oil_retention(self, metrics, oil_retention):
        """
        Evaluate metrics across oil retention and return an array with the all
        metric results.
        
        Parameters
        ----------
        metrics : Iterable[Callable; M elements]
            Should return a number given no parameters.
        oil_retention: array_like[P elements]
            Oil retentions to evaluate.
        
        Returns
        -------
        results : array[M x P]
            All metric results.
        
        """
        M = len(metrics)
        P = len(oil_retention)
        data = np.zeros([M, P])
        for i in range(P):
            self.load_oil_retention(oil_retention[i])
            self.system.simulate()
            data[:, i] = [j() for j in metrics]
        return data

    def evaluate_across_oil_content(self, 
            efficiency, oil_retention, 
            metrics, oil_content
        ):
        """
        Evaluate metrics at given efficiency and oil retention across a set of 
        oil contents. Return an array with the all metric results.
            
        Parameters
        ----------
        efficiency : float
            Oil extraction efficiency.
        oil_retention : float
            Fraction of oil retained in bagasse.
        metrics : Iterable[Callable; M elements]
            Should return a number given no parameters.
        oil_content : array_like[P elements]
            Oil content of feedstocks [dry wt. %].
        
        Returns
        -------
        results : array[shape x M x P]
            All metric results at given specifications.
        
        Notes
        -----
        This method is vectorized along titer and yield. If, for example,
        the parameters had the following dimensions:
            
        efficiency [Y x T], oil_retention [Y x T], metrics [M], oil_content [P]
        
        This method would return an array with the following dimensions:
        
        results [Y x T x M x P]
        
        """
        return evaluate_across_oil_content(self, 
                                             efficiency, oil_retention, 
                                             metrics, oil_content)
    def evaluate_across_oil_retention(self, 
            efficiency, oil_content, 
            metrics, oil_retention
        ):
        """
        Evaluate metrics at given efficiency and oil retention across a set of 
        oil contents. Return an array with the all metric results.
            
        Parameters
        ----------
        efficiency : float
            Oil extraction efficiency.
        oil_content : float
            Oil content of feedstocks [dry wt. %].
        metrics : Iterable[Callable; M elements]
            Should return a number given no parameters.
        oil_retention : array_like[P elements]
            Fraction of oil retained in bagasse.
        
        Returns
        -------
        results : array[shape x M x P]
            All metric results at given specifications.
        
        Notes
        -----
        This method is vectorized along efficiency and oil_content. If, for example,
        the parameters had the following dimensions:
            
        efficiency [Y x T], oil_content [Y x T], metrics [M], oil_retention [P]
        
        This method would return an array with the following dimensions:
        
        results [Y x T x M x P]
        
        """
        return evaluate_across_oil_retention(self, 
                                               efficiency, oil_content, 
                                               metrics, oil_retention)
